# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Progress Views
"""

from __future__ import unicode_literals, absolute_import

from tailbone.progress import get_progress_session


def progress(request):
    key = request.matchdict['key']
    session = get_progress_session(request, key, type=request.GET.get('sessiontype'))
    if session.get('complete'):
        msg = session.get('success_msg')
        if msg:
            request.session.flash(msg)
    elif session.get('error'):
        request.session.flash(session.get('error_msg', "An unspecified error occurred."), 'error')
    return session


def cancel(request):
    key = request.matchdict['key']
    session = get_progress_session(request, key)
    session.clear()
    session['canceled'] = True
    session.save()
    msg = request.params.get('cancel_msg', "The operation was canceled.")
    request.session.flash(msg)
    return {}


def includeme(config):
    config.add_route('progress', '/progress/{key}')
    config.add_view(progress, route_name='progress', renderer='json')

    config.add_route('progress.cancel', '/progress/{key}/cancel')
    config.add_view(cancel, route_name='progress.cancel', renderer='json')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for importer batches
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy as sa

from rattail.db import model

import colander

from tailbone.views.batch import BatchMasterView


class ImporterBatchView(BatchMasterView):
    """
    Master view for importer batches.
    """
    model_class = model.ImporterBatch
    default_handler_spec = 'rattail.batch.importer:ImporterBatchHandler'
    model_title_plural = "Import / Export Batches"
    route_prefix = 'batch.importer'
    url_prefix = '/batches/importer'
    template_prefix = '/batch/importer'
    creatable = False
    refreshable = False
    bulk_deletable = True
    rows_downloadable_csv = False
    rows_bulk_deletable = True

    labels = {
        'host_title': "Source",
        'local_title': "Target",
        'importer_key': "Model",
    }

    grid_columns = [
        'id',
        'description',
        'host_title',
        'local_title',
        'importer_key',
        'created',
        'created_by',
        'rowcount',
        'executed',
        'executed_by',
    ]

    form_fields = [
        'id',
        'description',
        'import_handler_spec',
        'host_title',
        'local_title',
        'importer_key',
        'notes',
        'created',
        'created_by',
        'row_table',
        'rowcount',
        'executed',
        'executed_by',
    ]

    row_grid_columns = [
        'sequence',
        'object_key',
        'object_str',
        'status_code',
    ]

    def configure_form(self, f):
        super(ImporterBatchView, self).configure_form(f)

        # readonly fields
        f.set_readonly('import_handler_spec')
        f.set_readonly('host_title')
        f.set_readonly('local_title')
        f.set_readonly('importer_key')
        f.set_readonly('row_table')

    def make_status_breakdown(self, batch):
        # TODO: should implement this, just can't use batch.data_rows apparently
        pass

    def delete_instance(self, batch):
        self.make_row_table(batch.row_table)
        if self.current_row_table is not None:
            self.current_row_table.drop()
        super(ImporterBatchView, self).delete_instance(batch)

    def make_row_table(self, name):
        if not hasattr(self, 'current_row_table'):
            metadata = sa.MetaData(schema='batch', bind=self.Session.bind)
            try:
                self.current_row_table = sa.Table(name, metadata, autoload=True)
            except sa.exc.NoSuchTableError:
                self.current_row_table = None

    def get_row_data(self, batch):
        self.make_row_table(batch.row_table)
        return self.Session.query(self.current_row_table)

    def get_row_status_enum(self):
        return self.enum.IMPORTER_BATCH_ROW_STATUS

    def configure_row_grid(self, g):
        super(ImporterBatchView, self).configure_row_grid(g)

        def make_filter(field, **kwargs):
            column = getattr(self.current_row_table.c, field)
            g.set_filter(field, column, **kwargs)

        make_filter('object_key')
        make_filter('object_str')
        make_filter('status_code', label="Status",
                    value_enum=self.enum.IMPORTER_BATCH_ROW_STATUS)

        def make_sorter(field):
            column = getattr(self.current_row_table.c, field)
            g.sorters[field] = lambda q, d: q.order_by(getattr(column, d)())

        make_sorter('sequence')
        make_sorter('object_key')
        make_sorter('object_str')
        make_sorter('status_code')

        g.set_sort_defaults('sequence')

        g.set_label('object_str', "Object Description")

        g.set_link('sequence')
        g.set_link('object_key')
        g.set_link('object_str')

    def row_grid_extra_class(self, row, i):
        if row.status_code == self.enum.IMPORTER_BATCH_ROW_STATUS_DELETE:
            return 'warning'
        if row.status_code in (self.enum.IMPORTER_BATCH_ROW_STATUS_CREATE,
                               self.enum.IMPORTER_BATCH_ROW_STATUS_UPDATE):
            return 'notice'

    def get_row_action_route_kwargs(self, row):
        return {
            'uuid': self.current_row_table.name,
            'row_uuid': row.uuid,
        }

    def get_row_instance(self):
        batch_uuid = self.request.matchdict['uuid']
        row_uuid = self.request.matchdict['row_uuid']
        self.make_row_table(batch_uuid)
        return self.Session.query(self.current_row_table)\
                           .filter(self.current_row_table.c.uuid == row_uuid)\
                           .one()

    def get_parent(self, row):
        uuid = self.current_row_table.name
        return self.Session.query(model.ImporterBatch).get(uuid)

    def get_row_instance_title(self, row):
        if row.object_str:
            return row.object_str
        if row.object_key:
            return row.object_key
        return "Row {}".format(row.sequence)

    def template_kwargs_view_row(self, **kwargs):
        batch = kwargs['parent_instance']
        row = kwargs['instance']
        kwargs['batch'] = batch
        kwargs['instance_title'] = batch.id_str

        fields = set()
        old_values = {}
        new_values = {}
        for col in self.current_row_table.c:
            if col.name.startswith('key_'):
                field = col.name[4:]
                fields.add(field)
                old_values[field] = new_values[field] = getattr(row, col.name)
            elif col.name.startswith('pre_'):
                field = col.name[4:]
                fields.add(field)
                old_values[field] = getattr(row, col.name)
            elif col.name.startswith('post_'):
                field = col.name[5:]
                fields.add(field)
                new_values[field] = getattr(row, col.name)

        kwargs['diff_fields'] = sorted(fields)
        kwargs['diff_old_values'] = old_values
        kwargs['diff_new_values'] = new_values
        return kwargs

    def make_row_form(self, instance=None, **kwargs):
        fields = ['sequence', 'object_key', 'object_str', 'status_code']
        for col in self.current_row_table.c:
            if col.name.startswith('key_'):
                fields.append(col.name)

        if kwargs.get('fields') is None:
            kwargs['fields'] = fields

        row = dict([(field, getattr(instance, field))
                    for field in fields])
        row['status_text'] = instance.status_text

        kwargs.setdefault('schema', colander.Schema())
        kwargs.setdefault('cancel_url', None)
        return super(ImporterBatchView, self).make_row_form(instance=row, **kwargs)

    def configure_row_form(self, f):
        """
        Configure the row form.
        """
        # object_str
        f.set_label('object_str', "Object Description")

        # status_code
        f.set_renderer('status_code', self.render_row_status_code)
        f.set_label('status_code', "Status")

    def render_row_status_code(self, row, field):
        status = self.enum.IMPORTER_BATCH_ROW_STATUS[row['status_code']]
        if row['status_code'] == self.enum.IMPORTER_BATCH_ROW_STATUS_UPDATE and row['status_text']:
            return "{} ({})".format(status, row['status_text'])
        return status

    def delete_row(self):
        row = self.get_row_instance()
        if not row:
            raise self.notfound()

        batch = self.get_parent(row)
        query = self.current_row_table.delete().where(self.current_row_table.c.uuid == row.uuid)
        query.execute()
        batch.rowcount -= 1
        return self.redirect(self.get_action_url('view', batch))

    def bulk_delete_rows(self):
        batch = self.get_instance()
        query = self.get_effective_row_data(sort=False)
        batch.rowcount -= query.count()
        delete_query = self.current_row_table.delete().where(self.current_row_table.c.uuid.in_([row.uuid for row in query]))
        delete_query.execute()
        return self.redirect(self.get_action_url('view', batch))


def includeme(config):
    ImporterBatchView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Tailbone Web API - Core Views
"""

from __future__ import unicode_literals, absolute_import

from tailbone.views import View


def api(view_meth):
    """
    Common decorator for all API views.  Ideally this would not be needed..but
    for now, alas, it is.
    """
    def wrapped(view, *args, **kwargs):

        # TODO: why doesn't this work here...? (instead we have to repeat this
        # code in lots of other places)
        # if view.request.method == 'OPTIONS':
        #     return view.request.response

        # invoke the view logic first, since presumably it may involve a
        # redirect in which case we don't really need to add the CSRF token.
        # main known use case for this is the /logout endpoint - if that gets
        # hit then the "current" (old) session will be destroyed, in which case
        # we can't use the token from that, but instead must generate a new one.
        result = view_meth(view, *args, **kwargs)

        # explicitly set CSRF token cookie, unless OPTIONS request 
        # TODO: why doesn't pyramid do this for us again?
        if view.request.method != 'OPTIONS':
            view.request.response.set_cookie(name='XSRF-TOKEN',
                                             value=view.request.session.get_csrf_token())

        return result

    return wrapped


class APIView(View):
    """
    Base class for all API views.
    """



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail config extension for Tailbone
"""

from __future__ import unicode_literals, absolute_import

from rattail.config import ConfigExtension as BaseExtension
from rattail.db.config import configure_session

from tailbone.db import Session


class ConfigExtension(BaseExtension):
    """
    Rattail config extension for Tailbone.  Does the following:

     * Adds the rattail config object to the constructor kwargs for the
       underlying Session factory.

     * Configures the main Tailbone database session so that it records
       changes, if the config file so dictates.
    """
    key = 'tailbone'

    def configure(self, config):
        Session.configure(rattail_config=config)
        configure_session(config, Session)


def expose_vuejs_experiments(config):
    return config.getbool('tailbone', 'expose_vuejs_experiments',
                          default=False)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Mobile Grids
"""

from __future__ import unicode_literals, absolute_import

from pyramid.renderers import render

from .core import Grid


class MobileGrid(Grid):
    """
    Base class for all mobile grids
    """

    def render_filters(self, template='/mobile/grids/filters_simple.mako', **kwargs):
        context = kwargs
        context['request'] = self.request
        context['grid'] = self
        return render(template, context)

    def render_grid(self, template='/mobile/grids/grid.mako', **kwargs):
        context = kwargs
        context['grid'] = self
        return render(template, context)

    def render_complete(self, template='/mobile/grids/complete.mako', **kwargs):
        context = kwargs
        context['grid'] = self
        return render(template, context)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Base views for time sheets
"""

from __future__ import unicode_literals, absolute_import

import datetime

import six
import sqlalchemy as sa

from rattail import enum
from rattail.db import model, api
from rattail.time import localtime, make_utc, get_sunday
from rattail.util import pretty_hours, hours_as_decimal

import colander
from deform import widget as dfwidget
from webhelpers2.html import tags, HTML

from tailbone import forms
from tailbone.db import Session
from tailbone.views import View


class ShiftFilter(colander.Schema):

    store = colander.SchemaNode(forms.types.StoreType())

    department = colander.SchemaNode(forms.types.DepartmentType())

    date = colander.SchemaNode(colander.Date())


class EmployeeShiftFilter(colander.Schema):

    employee = colander.SchemaNode(forms.types.EmployeeType())

    date = colander.SchemaNode(colander.Date())


class TimeSheetView(View):
    """
    Base view for time sheets.
    """
    key = None
    title = None
    model_class = None
    expose_employee_views = True

    # Set this to False to avoid the default behavior of auto-filtering by
    # current store.
    default_filter_store = True

    @classmethod
    def get_title(cls):
        return cls.title or cls.key.capitalize()

    @classmethod
    def get_url_prefix(cls):
        return getattr(cls, 'url_prefix', cls.key).rstrip('/')

    def get_timesheet_context(self):
        """
        Determine date/store/dept context from user's session and/or defaults.
        """
        date = None
        date_key = 'timesheet.{}.date'.format(self.key)
        if date_key in self.request.session:
            date_value = self.request.session.get(date_key)
            if date_value:
                try:
                    date = datetime.datetime.strptime(date_value, '%m/%d/%Y').date()
                except ValueError:
                    pass
        if not date:
            date = localtime(self.rattail_config).date()

        store = None
        department = None
        store_key = 'timesheet.{}.store'.format(self.key)
        department_key = 'timesheet.{}.department'.format(self.key)
        if store_key in self.request.session or department_key in self.request.session:
            store_uuid = self.request.session.get(store_key)
            if store_uuid:
                store = Session.query(model.Store).get(store_uuid) if store_uuid else None
            department_uuid = self.request.session.get(department_key)
            if department_uuid:
                department = Session.query(model.Department).get(department_uuid)
        else: # no store/department in session
            if self.default_filter_store:
                store = self.rattail_config.get('rattail', 'store')
                if store:
                    store = api.get_store(Session(), store)

        employees = Session.query(model.Employee)\
                           .filter(model.Employee.status == enum.EMPLOYEE_STATUS_CURRENT)
        if store:
            employees = employees.join(model.EmployeeStore)\
                                 .filter(model.EmployeeStore.store == store)
        if department:
            employees = employees.join(model.EmployeeDepartment)\
                                 .filter(model.EmployeeDepartment.department == department)

        return {
            'date': date,
            'store': store,
            'department': department,
            'employees': employees.all(),
        }

    def get_employee_context(self):
        """
        Determine employee/date context from user's session and/or defaults
        """
        date = None
        date_key = 'timesheet.{}.employee.date'.format(self.key)
        if date_key in self.request.session:
            date_value = self.request.session.get(date_key)
            if date_value:
                try:
                    date = datetime.datetime.strptime(date_value, '%m/%d/%Y').date()
                except ValueError:
                    pass
        if not date:
            date = localtime(self.rattail_config).date()

        employee = None
        employee_key = 'timesheet.{}.employee'.format(self.key)
        if employee_key in self.request.session:
            employee_uuid = self.request.session[employee_key]
            employee = Session.query(model.Employee).get(employee_uuid) if employee_uuid else None
        if not employee:
            employee = self.request.user.employee

        # force current user if not allowed to view all data
        if not self.request.has_perm('{}.viewall'.format(self.key)):
            employee = self.request.user.employee
            assert employee

        return {'date': date, 'employee': employee}

    def process_filter_form(self, form):
        """
        Process a "shift filter" form if one was in fact POST'ed.  If it was
        then we store new context in session and redirect to display as normal.
        """
        if form.validate(newstyle=True):
            store = form.validated['store']
            self.request.session['timesheet.{}.store'.format(self.key)] = store.uuid if store else None
            department = form.validated['department']
            self.request.session['timesheet.{}.department'.format(self.key)] = department.uuid if department else None
            date = form.validated['date']
            self.request.session['timesheet.{}.date'.format(self.key)] = date.strftime('%m/%d/%Y') if date else None
            raise self.redirect(self.request.current_route_url())

    def process_employee_filter_form(self, form):
        """
        Process an "employee shift filter" form if one was in fact POST'ed.  If it
        was then we store new context in session and redirect to display as normal.
        """
        if form.validate(newstyle=True):
            employee = form.validated['employee']
            self.request.session['timesheet.{}.employee'.format(self.key)] = employee.uuid if employee else None
            date = form.validated['date']
            self.request.session['timesheet.{}.employee.date'.format(self.key)] = date.strftime('%m/%d/%Y') if date else None
            raise self.redirect(self.request.current_route_url())

    def make_full_filter_form(self, context):
        form = forms.Form(schema=ShiftFilter(), request=self.request)

        stores = self.get_stores()
        store_values = [(s.uuid, "{} - {}".format(s.id, s.name)) for s in stores]
        store_values.insert(0, ('', "(all)"))
        form.set_widget('store', forms.widgets.PlainSelectWidget(values=store_values))
        if context['store']:
            form.set_default('store', context['store'].uuid)
        else:
            # TODO: why is this necessary? somehow the previous store is being
            # preserved as the "default" when switching from single store view
            # to "all stores" view
            form.set_default('store', '')

        departments = self.get_departments()
        department_values = [(d.uuid, d.name) for d in departments]
        department_values.insert(0, ('', "(all)"))
        form.set_widget('department', forms.widgets.PlainSelectWidget(values=department_values))
        if context['department']:
            form.set_default('department', context['department'].uuid)
        else:
            # TODO: why is this necessary? somehow the previous dept is being
            # preserved as the "default" when switching from single dept view
            # to "all depts" view
            form.set_default('department', '')

        form.set_type('date', 'date_jquery')
        form.set_default('date', get_sunday(context['date']))
        return form

    def full(self):
        """
        View a "full" timesheet/schedule, i.e. all employees but filterable by
        store and/or department.
        """
        context = self.get_timesheet_context()
        form = self.make_full_filter_form(context)
        self.process_filter_form(form)
        context['form'] = form
        return self.render_full(**context)

    def make_employee_filter_form(self, context):
        """
        View time sheet for single employee.
        """
        permission_prefix = self.key
        form = forms.Form(schema=EmployeeShiftFilter(), request=self.request)

        if self.request.has_perm('{}.viewall'.format(permission_prefix)):
            employee_display = six.text_type(context['employee'] or '')
            employees_url = self.request.route_url('employees.autocomplete')
            form.set_widget('employee', forms.widgets.JQueryAutocompleteWidget(
                field_display=employee_display, service_url=employees_url))
            if context['employee']:
                form.set_default('employee', context['employee'].uuid)
        else:
            form.set_widget('employee', forms.widgets.ReadonlyWidget())
            form.set_default('employee', context['employee'].uuid)

        form.set_type('date', 'date_jquery')
        form.set_default('date', get_sunday(context['date']))
        return form

    def employee(self):
        """
        View time sheet for single employee.
        """
        context = self.get_employee_context()
        form = self.make_employee_filter_form(context)
        self.process_employee_filter_form(form)
        context['form'] = form
        return self.render_single(**context)

    def crossview(self):
        """
        Update session storage to so 'other' view reflects current view
        filters, then redirect to other view.
        """
        other_key = 'timesheet' if self.key == 'schedule' else 'schedule'

        # TODO: this check is pretty hacky..
        # employee time sheet
        if 'employee' in self.request.get_referrer():
            self.session_put('employee', self.session_get('employee'), mainkey=other_key)
            self.session_put('employee.date', self.session_get('employee.date'), mainkey=other_key)
            return self.redirect(self.request.route_url('{}.employee'.format(other_key)))

        else: # full time sheet
            self.session_put('store', self.session_get('store'), mainkey=other_key)
            self.session_put('department', self.session_get('department'), mainkey=other_key)
            self.session_put('date', self.session_get('date'), mainkey=other_key)
            return self.redirect(self.request.route_url(other_key))

    def session_get(self, key, mainkey=None):
        if mainkey is None:
            mainkey = self.key
        return self.request.session.get('timesheet.{}.{}'.format(mainkey, key))

    def session_put(self, key, value, mainkey=None):
        if mainkey is None:
            mainkey = self.key
        self.request.session['timesheet.{}.{}'.format(mainkey, key)] = value

    def get_stores(self):
        return Session.query(model.Store).order_by(model.Store.id).all()

    def get_store_options(self, stores):
        options = [tags.Option("{} - {}".format(s.id, s.name), s.uuid) for s in stores]
        return tags.Options(options, prompt="(all)")

    def get_departments(self):
        return Session.query(model.Department).order_by(model.Department.name).all()

    def get_department_options(self, departments):
        options = [tags.Option(d.name, d.uuid) for d in departments]
        return tags.Options(options, prompt="(all)")

    def render_full(self, date=None, employees=None, store=None, department=None, form=None, **kwargs):
        """
        Render a time sheet for one or more employees, for the week which
        includes the specified date.
        """
        sunday = get_sunday(date)
        weekdays = [sunday]
        for i in range(1, 7):
            weekdays.append(sunday + datetime.timedelta(days=i))

        saturday = weekdays[-1]
        if saturday.year == sunday.year:
            week_of = '{} - {}'.format(sunday.strftime('%a %b %d'), saturday.strftime('%a %b %d, %Y'))
        else:
            week_of = '{} - {}'.format(sunday.strftime('%a %b %d, %Y'), saturday.strftime('%a %b %d, %Y'))

        self.modify_employees(employees, weekdays)

        stores = self.get_stores()
        store_options = self.get_store_options(stores)

        departments = self.get_departments()
        department_options = self.get_department_options(departments)

        context = {
            'page_title': self.get_title_full(),
            'form': form,
            'dform': form.make_deform_form() if form else None,
            'employees': employees,
            'stores': stores,
            'store_options': store_options,
            'store': store,
            'departments': departments,
            'department_options': department_options,
            'department': department,
            'week_of': week_of,
            'sunday': sunday,
            'prev_sunday': sunday - datetime.timedelta(days=7),
            'next_sunday': sunday + datetime.timedelta(days=7),
            'weekdays': weekdays,
            'permission_prefix': self.key,
            'render_shift': self.render_shift,
        }
        context.update(kwargs)
        return context

    def get_title_full(self):
        return "Full {}".format(self.get_title())

    def render_shift(self, shift):
        return HTML.tag('span', c=shift.get_display(self.rattail_config))

    def render_single(self, date=None, employee=None, form=None, **kwargs):
        """
        Render a time sheet for one employee, for the week which includes the
        specified date.
        """
        sunday = get_sunday(date)
        weekdays = [sunday]
        for i in range(1, 7):
            weekdays.append(sunday + datetime.timedelta(days=i))

        saturday = weekdays[-1]
        if saturday.year == sunday.year:
            week_of = '{} - {}'.format(sunday.strftime('%a %b %d'), saturday.strftime('%a %b %d, %Y'))
        else:
            week_of = '{} - {}'.format(sunday.strftime('%a %b %d, %Y'), saturday.strftime('%a %b %d, %Y'))

        self.modify_employees([employee], weekdays)

        context = {
            'single': True,
            'page_title': "Employee {}".format(self.get_title()),
            'form': form,
            'dform': form.make_deform_form() if form else None,
            'employee': employee,
            'employees': [employee],
            'week_of': week_of,
            'sunday': sunday,
            'prev_sunday': sunday - datetime.timedelta(days=7),
            'next_sunday': sunday + datetime.timedelta(days=7),
            'weekdays': weekdays,
            'permission_prefix': self.key,
            'render_shift': self.render_shift,
        }
        context.update(kwargs)
        return context

    def modify_employees(self, employees, weekdays):
        self.fetch_shift_data(self.model_class, employees, weekdays)

    def fetch_shift_data(self, cls, employees, weekdays):
        """
        Fetch all shift data of the given model class (``cls``), according to
        the given params.  The cached shift data is attached to each employee.
        """
        # TODO: a bit hacky, this?  display hours as HH:MM by default, but
        # check config in order to display as HH.HH for certain users
        hours_style = 'pretty'
        if self.request.user:
            hours_style = self.rattail_config.get('tailbone', 'hours_style.{}'.format(self.request.user.username),
                                                  default='pretty')
        if hours_style != 'decimal':
            hours_style = 'pretty'

        shift_type = 'scheduled' if cls is model.ScheduledShift else 'worked'
        min_time = localtime(self.rattail_config, datetime.datetime.combine(weekdays[0], datetime.time(0)))
        max_time = localtime(self.rattail_config, datetime.datetime.combine(weekdays[-1] + datetime.timedelta(days=1), datetime.time(0)))
        shifts = Session.query(cls)\
                        .filter(cls.employee_uuid.in_([e.uuid for e in employees]))\
                        .filter(sa.or_(
                            sa.and_(
                                cls.start_time >= make_utc(min_time),
                                cls.start_time < make_utc(max_time),
                            ),
                            sa.and_(
                                cls.start_time == None,
                                cls.end_time >= make_utc(min_time),
                                cls.end_time < make_utc(max_time),
                            )))\
                        .all()

        for employee in employees:
            employee_shifts = sorted([s for s in shifts if s.employee_uuid == employee.uuid],
                                     key=lambda s: s.start_time or s.end_time)
            if not hasattr(employee, 'weekdays'):
                employee.weekdays = [{} for day in weekdays]
            setattr(employee, '{}_hours'.format(shift_type), datetime.timedelta(0))
            setattr(employee, '{}_hours_display'.format(shift_type), '0')
            hours_incomplete = False

            for i, day in enumerate(weekdays):
                empday = {
                    '{}_shifts'.format(shift_type): [],
                    '{}_hours'.format(shift_type): datetime.timedelta(0),
                    '{}_hours_display'.format(shift_type): '',
                    'hours_incomplete': False,
                }

                while employee_shifts:
                    shift = employee_shifts[0]
                    if shift.employee_uuid != employee.uuid:
                        break
                    elif shift.get_date(self.rattail_config) == day:
                        empday['{}_shifts'.format(shift_type)].append(shift)
                        length = shift.length
                        if length is not None:
                            empday['{}_hours'.format(shift_type)] += shift.length
                            setattr(employee, '{}_hours'.format(shift_type),
                                    getattr(employee, '{}_hours'.format(shift_type)) + shift.length)
                        else:
                            hours_incomplete = True
                            empday['hours_incomplete'] = True
                        del employee_shifts[0]
                    else:
                        break

                hours = empday['{}_hours'.format(shift_type)]
                if hours:
                    if hours_style == 'pretty':
                        display = pretty_hours(hours)
                    else: # decimal
                        display = six.text_type(hours_as_decimal(hours))
                    if empday['hours_incomplete']:
                        display = '{} ?'.format(display)
                    empday['{}_hours_display'.format(shift_type)] = display
                employee.weekdays[i].update(empday)

            hours = getattr(employee, '{}_hours'.format(shift_type))
            if hours:
                if hours_style == 'pretty':
                    display = pretty_hours(hours)
                else: # decimal
                    display = six.text_type(hours_as_decimal(hours))
                if hours_incomplete:
                    display = '{} ?'.format(display)
                setattr(employee, '{}_hours_display'.format(shift_type), display)

    @classmethod
    def defaults(cls, config):
        """
        Provide default configuration for a time sheet view.
        """
        cls._defaults(config)

    @classmethod
    def _defaults(cls, config):
        """
        Provide default configuration for a time sheet view.
        """
        title = cls.get_title()
        url_prefix = cls.get_url_prefix()
        config.add_tailbone_permission_group(cls.key, title)
        config.add_tailbone_permission(cls.key, '{}.viewall'.format(cls.key), "View full {}".format(title))

        # full time sheet
        config.add_route(cls.key, '{}/'.format(url_prefix))
        config.add_view(cls, attr='full', route_name=cls.key,
                        renderer='/shifts/{}.mako'.format(cls.key),
                        permission='{}.viewall'.format(cls.key))

        # single employee time sheet
        if cls.expose_employee_views:
            config.add_tailbone_permission(cls.key, '{}.view'.format(cls.key), "View single employee {}".format(title))
            config.add_route('{}.employee'.format(cls.key), '{}/employee/'.format(url_prefix))
            config.add_view(cls, attr='employee', route_name='{}.employee'.format(cls.key),
                            renderer='/shifts/{}.mako'.format(cls.key),
                            permission='{}.view'.format(cls.key))

        # goto cross-view (view 'timesheet' as 'schedule' or vice-versa)
        other_key = 'timesheet' if cls.key == 'schedule' else 'schedule'
        config.add_route('{}.goto.{}'.format(cls.key, other_key), '{}/goto-{}'.format(url_prefix, other_key))
        config.add_view(cls, attr='crossview', route_name='{}.goto.{}'.format(cls.key, other_key),
                        permission='{}.view'.format(other_key))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views with info about the underlying Rattail tables
"""

from __future__ import unicode_literals, absolute_import

from tailbone.views import MasterView


class TablesView(MasterView):
    """
    Master view for tables
    """
    normalized_model_name = 'table'
    model_key = 'name'
    model_title = "Table"
    creatable = False
    editable = False
    deletable = False
    viewable = False
    filterable = False
    pageable = False

    grid_columns = [
        'name',
        'row_count',
    ]

    def get_data(self, **kwargs):
        """
        Fetch existing table names and estimate row counts via PG SQL
        """
        sql = """
        select schemaname, relname, n_live_tup
        from pg_stat_user_tables
        order by n_live_tup desc;
        """
        result = self.Session.execute(sql)
        return [dict(name=row[1], row_count=row[2]) for row in result]

    def configure_grid(self, g):
        g.sorters['name'] = g.make_simple_sorter('name', foldcase=True)
        g.sorters['row_count'] = g.make_simple_sorter('row_count')
        g.set_sort_defaults('name')


def includeme(config):
    TablesView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Base View Class
"""

from __future__ import unicode_literals, absolute_import

import os

import six

from rattail.db import model
from rattail.util import progress_loop

from pyramid import httpexceptions
from pyramid.renderers import render_to_response
from pyramid.response import FileResponse

from tailbone.db import Session
from tailbone.auth import logout_user
from tailbone.util import should_use_buefy


class View(object):
    """
    Base class for all class-based views.
    """

    def __init__(self, request, context=None):
        self.request = request

        # if user becomes inactive while logged in, log them out
        if getattr(request, 'user', None):
            # TODO: why is the user sometimes not attached to session?
            # (this has only been seen on mobile, when creating a new ordering batch)
            if request.user not in Session():
                request.user = Session.merge(request.user)
            if not request.user.active:
                headers = logout_user(request)
                raise self.redirect(request.route_url('home'))

        config = self.rattail_config
        if config:
            self.enum = config.get_enum()

    @property
    def rattail_config(self):
        """
        Reference to the effective Rattail config object.
        """
        return getattr(self.request, 'rattail_config', None)

    def notfound(self):
        return httpexceptions.HTTPNotFound()
    
    def get_use_buefy(self):
        """
        Returns a flag indicating whether or not the current theme supports
        (and therefore should use) the Buefy JS library.
        """
        return should_use_buefy(self.request)

    def late_login_user(self):
        """
        Returns the :class:`rattail:rattail.db.model.User` instance
        corresponding to the "late login" form data (if any), or ``None``.
        """
        if self.request.method == 'POST':
            uuid = self.request.POST.get('late-login-user')
            if uuid:
                return Session.query(model.User).get(uuid)

    def redirect(self, url, **kwargs):
        """
        Convenience method to return a HTTP 302 response.
        """
        return httpexceptions.HTTPFound(location=url, **kwargs)

    def progress_loop(self, func, items, factory, *args, **kwargs):
        return progress_loop(func, items, factory, *args, **kwargs)

    # TODO: this signature seems wonky
    def render_progress(self, progress, kwargs, template=None):
        """
        Render the progress page, with given kwargs as context.
        """
        if not template:
            template = '/progress.mako'
        kwargs['progress'] = progress
        kwargs.setdefault('can_cancel', True)
        return render_to_response(template, kwargs, request=self.request)

    def json_response(self, data):
        """
        Convenience method to return a JSON response.
        """
        return render_to_response('json', data,
                                  request=self.request)

    def file_response(self, path):
        """
        Returns a generic FileResponse from the given path
        """
        if not os.path.exists(path):
            return self.notfound()
        response = FileResponse(path, request=self.request)
        response.content_length = os.path.getsize(path)
        filename = os.path.basename(path)
        if six.PY2:
            filename = filename.encode('ascii', 'replace')
        response.content_disposition = str('attachment; filename="{}"'.format(filename))
        return response



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Label Profile Views
"""

from __future__ import unicode_literals, absolute_import

from rattail.db import model

import colander

from tailbone import forms
from tailbone.views import MasterView


class ProfilesView(MasterView):
    """
    Master view for the LabelProfile model.
    """
    model_class = model.LabelProfile
    model_title = "Label Profile"
    url_prefix = '/labels/profiles'
    has_versions = True

    grid_columns = [
        'ordinal',
        'code',
        'description',
        'visible',
        'sync_me',
    ]

    form_fields = [
        'ordinal',
        'code',
        'description',
        'printer_spec',
        'formatter_spec',
        'format',
        'visible',
        'sync_me',
    ]

    def configure_grid(self, g):
        super(ProfilesView, self).configure_grid(g)
        g.set_sort_defaults('ordinal')
        g.set_type('visible', 'boolean')
        g.set_link('code')
        g.set_link('description')

    def configure_form(self, f):
        super(ProfilesView, self).configure_form(f)

        # format
        f.set_type('format', 'codeblock')

    def after_create(self, profile):
        self.after_edit(profile)

    def after_edit(self, profile):
        if not profile.format:
            formatter = profile.get_formatter(self.rattail_config)
            if formatter:
                try:
                    profile.format = formatter.default_format
                except NotImplementedError:
                    pass

    def make_printer_settings_form(self, profile, printer):
        use_buefy = self.get_use_buefy()
        schema = colander.Schema()

        for name, label in printer.required_settings.items():
            node = colander.SchemaNode(colander.String(),
                                       name=name,
                                       title=label,
                                       default=profile.get_printer_setting(name))
            schema.add(node)

        form = forms.Form(schema=schema, request=self.request,
                          use_buefy=use_buefy,
                          model_instance=profile,
                          # TODO: ugh, this is necessary to avoid some logic
                          # which assumes a ColanderAlchemy schema i think?
                          appstruct=None)
        form.cancel_url = self.get_action_url('view', profile)
        form.auto_disable_cancel = True

        form.insert_before(schema.children[0].name, 'label_profile')
        form.set_readonly('label_profile')
        form.set_renderer('label_profile', lambda p, f: p.description)

        form.insert_after('label_profile', 'printer_spec')
        form.set_readonly('printer_spec')
        form.set_renderer('printer_spec', lambda p, f: p.printer_spec)

        return form

    def printer_settings(self):
        """
        View for editing extended Printer Settings, for a given Label Profile.
        """
        profile = self.get_instance()
        read_profile = self.redirect(self.get_action_url('view', profile))

        printer = profile.get_printer(self.rattail_config)
        if not printer:
            msg = "Label profile \"{}\" does not have a functional printer spec.".format(profile)
            self.request.session.flash(msg)
            return read_profile
        if not printer.required_settings:
            msg = "Printer class for label profile \"{}\" does not require any settings.".format(profile)
            self.request.session.flash(msg)
            return read_profile

        form = self.make_printer_settings_form(profile, printer)

        # TODO: should use form.validate() here
        if self.request.method == 'POST':
            for setting in printer.required_settings:
                if setting in self.request.POST:
                    profile.save_printer_setting(setting, self.request.POST[setting])
            return read_profile

        return self.render_to_response('printer', {
            'form': form,
            'dform': form.make_deform_form(),
            'profile': profile,
            'printer': printer,
        })

    @classmethod
    def defaults(cls, config):
        cls._defaults(config)
        cls._labelprofile_defaults(config)

    @classmethod
    def _labelprofile_defaults(cls, config):
        route_prefix = cls.get_route_prefix()
        url_prefix = cls.get_url_prefix()
        permission_prefix = cls.get_permission_prefix()
        model_key = cls.get_model_key()

        # edit printer settings
        config.add_route('{}.printer_settings'.format(route_prefix), '{}/{{{}}}/printer'.format(url_prefix, model_key))
        config.add_view(cls, attr='printer_settings', route_name='{}.printer_settings'.format(route_prefix),
                        permission='{}.edit'.format(permission_prefix))


def includeme(config):
    ProfilesView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Core Grid Classes
"""

from __future__ import unicode_literals, absolute_import

import datetime
from six.moves import urllib

import six
import sqlalchemy as sa
from sqlalchemy import orm

from rattail.db import api
from rattail.db.types import GPCType
from rattail.util import prettify, pretty_boolean, pretty_quantity, pretty_hours
from rattail.time import localtime

import webhelpers2_grid
from pyramid.renderers import render
from webhelpers2.html import HTML, tags
from paginate_sqlalchemy import SqlalchemyOrmPage

from . import filters as gridfilters
from tailbone.db import Session
from tailbone.util import raw_datetime


class FieldList(list):
    """
    Convenience wrapper for a field list.
    """

    def insert_before(self, field, newfield):
        i = self.index(field)
        self.insert(i, newfield)

    def insert_after(self, field, newfield):
        i = self.index(field)
        self.insert(i + 1, newfield)


class Grid(object):
    """
    Core grid class.  In sore need of documentation.
    """

    def __init__(self, key, data, columns=None, width='auto', request=None, mobile=False,
                 model_class=None, model_title=None, model_title_plural=None,
                 enums={}, labels={}, renderers={}, extra_row_class=None, linked_columns=[], url='#',
                 joiners={}, filterable=False, filters={}, use_byte_string_filters=False,
                 sortable=False, sorters={}, default_sortkey=None, default_sortdir='asc',
                 pageable=False, default_pagesize=20, default_page=1,
                 checkboxes=False, checked=None, main_actions=[], more_actions=[], delete_speedbump=False,
                 **kwargs):

        self.key = key
        self.data = data
        self.columns = FieldList(columns) if columns is not None else None
        self.width = width
        self.request = request
        self.mobile = mobile
        self.model_class = model_class
        if self.model_class and self.columns is None:
            self.columns = self.make_columns()

        self.model_title = model_title
        if not self.model_title and self.model_class and hasattr(self.model_class, 'get_model_title'):
            self.model_title = self.model_class.get_model_title()

        self.model_title_plural = model_title_plural
        if not self.model_title_plural:
            if self.model_class and hasattr(self.model_class, 'get_model_title_plural'):
                self.model_title_plural = self.model_class.get_model_title_plural()
            if not self.model_title_plural:
                self.model_title_plural = '{}s'.format(self.model_title)

        self.enums = enums or {}

        self.labels = labels or {}
        self.renderers = self.make_default_renderers(renderers or {})
        self.extra_row_class = extra_row_class
        self.linked_columns = linked_columns or []
        self.url = url
        self.joiners = joiners or {}

        self.filterable = filterable
        self.use_byte_string_filters = use_byte_string_filters
        self.filters = self.make_filters(filters)

        self.sortable = sortable
        self.sorters = self.make_sorters(sorters)
        self.default_sortkey = default_sortkey
        self.default_sortdir = default_sortdir

        self.pageable = pageable
        self.default_pagesize = default_pagesize
        self.default_page = default_page

        self.checkboxes = checkboxes
        self.checked = checked
        if self.checked is None:
            self.checked = lambda item: False
        self.main_actions = main_actions or []
        self.more_actions = more_actions or []
        self.delete_speedbump = delete_speedbump

        self._whgrid_kwargs = kwargs

    def make_columns(self):
        """
        Return a default list of columns, based on :attr:`model_class`.
        """
        if not self.model_class:
            raise ValueError("Must define model_class to use make_columns()")

        mapper = orm.class_mapper(self.model_class)
        return [prop.key for prop in mapper.iterate_properties]

    def hide_column(self, key):
        if key in self.columns:
            self.columns.remove(key)

    def hide_columns(self, *keys):
        for key in keys:
            self.hide_column(key)

    def append(self, field):
        self.columns.append(field)

    def insert_before(self, field, newfield):
        self.columns.insert_before(field, newfield)

    def insert_after(self, field, newfield):
        self.columns.insert_after(field, newfield)

    def replace(self, oldfield, newfield):
        self.insert_after(oldfield, newfield)
        self.hide_column(oldfield)

    def set_joiner(self, key, joiner):
        if joiner is None:
            self.joiners.pop(key, None)
        else:
            self.joiners[key] = joiner

    def set_sorter(self, key, *args, **kwargs):
        self.sorters[key] = self.make_sorter(*args, **kwargs)

    def set_sort_defaults(self, sortkey, sortdir='asc'):
        self.default_sortkey = sortkey
        self.default_sortdir = sortdir

    def set_filter(self, key, *args, **kwargs):
        if len(args) == 1 and args[0] is None:
            self.remove_filter(key)
        else:
            if 'label' not in kwargs and key in self.labels:
                kwargs['label'] = self.labels[key]
            self.filters[key] = self.make_filter(key, *args, **kwargs)

    def remove_filter(self, key):
        self.filters.pop(key, None)

    def set_label(self, key, label):
        self.labels[key] = label
        if key in self.filters:
            self.filters[key].label = label

    def get_label(self, key):
        """
        Returns the label text for given field key.
        """
        return self.labels.get(key, prettify(key))

    def set_link(self, key, link=True):
        if link:
            if key not in self.linked_columns:
                self.linked_columns.append(key)
        else: # unlink
            if self.linked_columns and key in self.linked_columns:
                self.linked_columns.remove(key)

    def set_renderer(self, key, renderer):
        self.renderers[key] = renderer

    def set_type(self, key, type_):
        if type_ == 'boolean':
            self.set_renderer(key, self.render_boolean)
        elif type_ == 'currency':
            self.set_renderer(key, self.render_currency)
        elif type_ == 'datetime':
            self.set_renderer(key, self.render_datetime)
        elif type_ == 'datetime_local':
            self.set_renderer(key, self.render_datetime_local)
        elif type_ == 'enum':
            self.set_renderer(key, self.render_enum)
        elif type_ == 'gpc':
            self.set_renderer(key, self.render_gpc)
        elif type_ == 'percent':
            self.set_renderer(key, self.render_percent)
        elif type_ == 'quantity':
            self.set_renderer(key, self.render_quantity)
        elif type_ == 'duration':
            self.set_renderer(key, self.render_duration)
        elif type_ == 'duration_hours':
            self.set_renderer(key, self.render_duration_hours)
        else:
            raise ValueError("Unsupported type for column '{}': {}".format(key, type_))

    def set_enum(self, key, enum):
        if enum:
            self.enums[key] = enum
            self.set_type(key, 'enum')
            if key in self.filters:
                self.filters[key].set_value_renderer(gridfilters.EnumValueRenderer(enum))
        else:
            self.enums.pop(key, None)

    def render_generic(self, obj, column_name):
        return self.obtain_value(obj, column_name)

    def render_boolean(self, obj, column_name):
        value = self.obtain_value(obj, column_name)
        return pretty_boolean(value)

    def obtain_value(self, obj, column_name):
        try:
            return obj[column_name]
        except TypeError:
            return getattr(obj, column_name)

    def render_currency(self, obj, column_name):
        value = self.obtain_value(obj, column_name)
        if value is None:
            return ""
        if value < 0:
            return "(${:0,.2f})".format(0 - value)
        return "${:0,.2f}".format(value)

    def render_datetime(self, obj, column_name):
        value = self.obtain_value(obj, column_name)
        if value is None:
            return ""
        return raw_datetime(self.request.rattail_config, value)

    def render_datetime_local(self, obj, column_name):
        value = self.obtain_value(obj, column_name)
        if value is None:
            return ""
        value = localtime(self.request.rattail_config, value)
        return raw_datetime(self.request.rattail_config, value)

    def render_enum(self, obj, column_name):
        value = self.obtain_value(obj, column_name)
        if value is None:
            return ""
        enum = self.enums.get(column_name)
        if enum and value in enum:
            return six.text_type(enum[value])
        return six.text_type(value)

    def render_gpc(self, obj, column_name):
        value = self.obtain_value(obj, column_name)
        if value is None:
            return ""
        return value.pretty()

    def render_percent(self, obj, column_name):
        value = self.obtain_value(obj, column_name)
        if value is None:
            return ""
        return "{:0.3f} %".format(value * 100)

    def render_quantity(self, obj, column_name):
        value = self.obtain_value(obj, column_name)
        return pretty_quantity(value)

    def render_duration(self, obj, column_name):
        value = self.obtain_value(obj, column_name)
        if value is None:
            return ""
        return pretty_hours(datetime.timedelta(seconds=value))

    def render_duration_hours(self, obj, field):
        value = self.obtain_value(obj, field)
        if value is None:
            return ""
        return pretty_hours(hours=value)

    def set_url(self, url):
        self.url = url

    def make_url(self, obj, i=None):
        if callable(self.url):
            return self.url(obj)
        return self.url

    def make_webhelpers_grid(self):
        kwargs = dict(self._whgrid_kwargs)
        kwargs['request'] = self.request
        kwargs['mobile'] = self.mobile
        kwargs['url'] = self.make_url

        columns = list(self.columns)
        column_labels = kwargs.setdefault('column_labels', {})
        column_formats = kwargs.setdefault('column_formats', {})

        for key, value in self.labels.items():
            column_labels.setdefault(key, value)

        if self.checkboxes:
            columns.insert(0, 'checkbox')
            column_labels['checkbox'] = tags.checkbox('check-all')
            column_formats['checkbox'] = self.checkbox_column_format

        if self.renderers:
            kwargs['renderers'] = self.renderers
        if self.extra_row_class:
            kwargs['extra_record_class'] = self.extra_row_class
        if self.linked_columns:
            kwargs['linked_columns'] = list(self.linked_columns)

        if self.main_actions or self.more_actions:
            columns.append('actions')
            column_formats['actions'] = self.actions_column_format

        # TODO: pretty sure this factory doesn't serve all use cases yet?
        factory = CustomWebhelpersGrid
        # factory = webhelpers2_grid.Grid
        if self.sortable:
            # factory = CustomWebhelpersGrid
            kwargs['order_column'] = self.sortkey
            kwargs['order_direction'] = 'dsc' if self.sortdir == 'desc' else 'asc'

        grid = factory(self.make_visible_data(), columns, **kwargs)
        if self.sortable:
            grid.exclude_ordering = list([key for key in grid.exclude_ordering
                                          if key not in self.sorters])
        return grid

    def make_default_renderers(self, renderers):
        """
        Make the default set of column renderers for the grid.

        We honor any existing renderers which have already been set, but then
        we also try to supplement that by auto-assigning renderers based on
        underlying column type.  Note that this special logic only applies to
        grids with a valid :attr:`model_class`.
        """
        if self.model_class:
            mapper = orm.class_mapper(self.model_class)
            for prop in mapper.iterate_properties:
                if isinstance(prop, orm.ColumnProperty) and not prop.key.endswith('uuid'):
                    if prop.key in self.columns and prop.key not in renderers:
                        if len(prop.columns) == 1:
                            coltype = prop.columns[0].type
                            renderers[prop.key] = self.get_renderer_for_column_type(coltype)

        return renderers

    def get_renderer_for_column_type(self, coltype):
        """
        Returns an appropriate renderer according to the given SA column type.
        """
        if isinstance(coltype, sa.Boolean):
            return self.render_boolean

        if isinstance(coltype, sa.DateTime):
            return self.render_datetime

        if isinstance(coltype, GPCType):
            return self.render_gpc

        return self.render_generic

    def checkbox_column_format(self, column_number, row_number, item):
        return HTML.td(self.render_checkbox(item), class_='checkbox')

    def actions_column_format(self, column_number, row_number, item):
        return HTML.td(self.render_actions(item, row_number), class_='actions')

    def render_grid(self, template='/grids/grid.mako', **kwargs):
        context = kwargs
        context['grid'] = self
        context['request'] = self.request
        grid_class = ''
        if self.width == 'full':
            grid_class = 'full'
        elif self.width == 'half':
            grid_class = 'half'
        context['grid_class'] = '{} {}'.format(grid_class, context.get('grid_class', ''))
        context.setdefault('grid_attrs', {})
        return render(template, context)

    def get_default_filters(self):
        """
        Returns the default set of filters provided by the grid.
        """
        if hasattr(self, 'default_filters'):
            if callable(self.default_filters):
                return self.default_filters()
            return self.default_filters
        filters = gridfilters.GridFilterSet()
        if self.model_class:
            mapper = orm.class_mapper(self.model_class)
            for prop in mapper.iterate_properties:
                if not isinstance(prop, orm.ColumnProperty):
                    continue
                if prop.key.endswith('uuid'):
                    continue
                if len(prop.columns) != 1:
                    continue
                column = prop.columns[0]
                if isinstance(column.type, sa.LargeBinary):
                    continue
                filters[prop.key] = self.make_filter(prop.key, column)
        return filters

    def make_filters(self, filters=None):
        """
        Returns an initial set of filters which will be available to the grid.
        The grid itself may or may not provide some default filters, and the
        ``filters`` kwarg may contain additions and/or overrides.
        """
        if filters:
            return filters
        return self.get_default_filters()

    def make_filter(self, key, column, **kwargs):
        """
        Make a filter suitable for use with the given column.
        """
        factory = kwargs.pop('factory', None)
        if not factory:
            factory = gridfilters.AlchemyGridFilter
            if isinstance(column.type, sa.String):
                factory = gridfilters.AlchemyStringFilter
            elif isinstance(column.type, sa.Numeric):
                factory = gridfilters.AlchemyNumericFilter
            elif isinstance(column.type, sa.Integer):
                factory = gridfilters.AlchemyIntegerFilter
            elif isinstance(column.type, sa.Boolean):
                # TODO: check column for nullable here?
                factory = gridfilters.AlchemyNullableBooleanFilter
            elif isinstance(column.type, sa.Date):
                factory = gridfilters.AlchemyDateFilter
            elif isinstance(column.type, sa.DateTime):
                factory = gridfilters.AlchemyDateTimeFilter
            elif isinstance(column.type, GPCType):
                factory = gridfilters.AlchemyGPCFilter
        kwargs.setdefault('encode_values', self.use_byte_string_filters)
        return factory(key, column=column, config=self.request.rattail_config, **kwargs)

    def iter_filters(self):
        """
        Iterate over all filters available to the grid.
        """
        return six.itervalues(self.filters)

    def iter_active_filters(self):
        """
        Iterate over all *active* filters for the grid.  Whether a filter is
        active is determined by current grid settings.
        """
        for filtr in self.iter_filters():
            if filtr.active:
                yield filtr

    def make_sorters(self, sorters=None):
        """
        Returns an initial set of sorters which will be available to the grid.
        The grid itself may or may not provide some default sorters, and the
        ``sorters`` kwarg may contain additions and/or overrides.
        """
        sorters, updates = {}, sorters
        if self.model_class:
            mapper = orm.class_mapper(self.model_class)
            for prop in mapper.iterate_properties:
                if isinstance(prop, orm.ColumnProperty) and not prop.key.endswith('uuid'):
                    sorters[prop.key] = self.make_sorter(prop)
        if updates:
            sorters.update(updates)
        return sorters

    def make_sorter(self, model_property):
        """
        Returns a function suitable for a sort map callable, with typical logic
        built in for sorting applied to ``field``.
        """
        class_ = getattr(model_property, 'class_', self.model_class)
        column = getattr(class_, model_property.key)
        return lambda q, d: q.order_by(getattr(column, d)())

    def make_simple_sorter(self, key, foldcase=False):
        """
        Returns a function suitable for a sort map callable, with typical logic
        built in for sorting a data set comprised of dicts, on the given key.
        """
        if foldcase:
            keyfunc = lambda v: v[key].lower()
        else:
            keyfunc = lambda v: v[key]
        return lambda q, d: sorted(q, key=keyfunc, reverse=d == 'desc')

    def load_settings(self, store=True):
        """
        Load current/effective settings for the grid, from the request query
        string and/or session storage.  If ``store`` is true, then once
        settings have been fully read, they are stored in current session for
        next time.  Finally, various instance attributes of the grid and its
        filters are updated in-place to reflect the settings; this is so code
        needn't access the settings dict directly, but the more Pythonic
        instance attributes.
        """

        # initial default settings
        settings = {}
        if self.sortable:
            settings['sortkey'] = self.default_sortkey
            settings['sortdir'] = self.default_sortdir
        if self.pageable:
            settings['pagesize'] = self.default_pagesize
            settings['page'] = self.default_page
        if self.filterable:
            for filtr in self.iter_filters():
                settings['filter.{}.active'.format(filtr.key)] = filtr.default_active
                settings['filter.{}.verb'.format(filtr.key)] = filtr.default_verb
                settings['filter.{}.value'.format(filtr.key)] = filtr.default_value

        # If user has default settings on file, apply those first.
        if self.user_has_defaults():
            self.apply_user_defaults(settings)

        # If request contains instruction to reset to default filters, then we
        # can skip the rest of the request/session checks.
        if self.request.GET.get('reset-to-default-filters') == 'true':
            pass

        # If request has filter settings, grab those, then grab sort/pager
        # settings from request or session.
        elif self.filterable and self.request_has_settings('filter'):
            self.update_filter_settings(settings, 'request')
            if self.request_has_settings('sort'):
                self.update_sort_settings(settings, 'request')
            else:
                self.update_sort_settings(settings, 'session')
            self.update_page_settings(settings)

        # If request has no filter settings but does have sort settings, grab
        # those, then grab filter settings from session, then grab pager
        # settings from request or session.
        elif self.request_has_settings('sort'):
            self.update_sort_settings(settings, 'request')
            self.update_filter_settings(settings, 'session')
            self.update_page_settings(settings)

        # NOTE: These next two are functionally equivalent, but are kept
        # separate to maintain the narrative...

        # If request has no filter/sort settings but does have pager settings,
        # grab those, then grab filter/sort settings from session.
        elif self.request_has_settings('page'):
            self.update_page_settings(settings)
            self.update_filter_settings(settings, 'session')
            self.update_sort_settings(settings, 'session')

        # If request has no settings, grab all from session.
        elif self.session_has_settings():
            self.update_filter_settings(settings, 'session')
            self.update_sort_settings(settings, 'session')
            self.update_page_settings(settings)

        # If no settings were found in request or session, don't store result.
        else:
            store = False
            
        # Maybe store settings for next time.
        if store:
            self.persist_settings(settings, 'session')

        # If request contained instruction to save current settings as defaults
        # for the current user, then do that.
        if self.request.GET.get('save-current-filters-as-defaults') == 'true':
            self.persist_settings(settings, 'defaults')

        # update ourself to reflect settings
        if self.filterable:
            for filtr in self.iter_filters():
                filtr.active = settings['filter.{}.active'.format(filtr.key)]
                filtr.verb = settings['filter.{}.verb'.format(filtr.key)]
                filtr.value = settings['filter.{}.value'.format(filtr.key)]
        if self.sortable:
            self.sortkey = settings['sortkey']
            self.sortdir = settings['sortdir']
        if self.pageable:
            self.pagesize = settings['pagesize']
            self.page = settings['page']

    def user_has_defaults(self):
        """
        Check to see if the current user has default settings on file for this grid.
        """
        user = self.request.user
        if not user:
            return False

        # NOTE: we used to leverage `self.session` here, but sometimes we might
        # be showing a grid of data from another system...so always use
        # Tailbone Session now, for the settings.  hopefully that didn't break
        # anything...
        session = Session()
        if user not in session:
            user = session.merge(user)

        # User defaults should have all or nothing, so just check one key.
        key = 'tailbone.{}.grid.{}.sortkey'.format(user.uuid, self.key)
        return api.get_setting(session, key) is not None

    def apply_user_defaults(self, settings):
        """
        Update the given settings dict with user defaults, if any exist.
        """
        def merge(key, normalize=lambda v: v):
            skey = 'tailbone.{}.grid.{}.{}'.format(self.request.user.uuid, self.key, key)
            value = api.get_setting(Session(), skey)
            settings[key] = normalize(value)

        if self.filterable:
            for filtr in self.iter_filters():
                merge('filter.{}.active'.format(filtr.key), lambda v: v == 'true')
                merge('filter.{}.verb'.format(filtr.key))
                merge('filter.{}.value'.format(filtr.key))

        if self.sortable:
            merge('sortkey')
            merge('sortdir')

        if self.pageable:
            merge('pagesize', int)
            merge('page', int)

    def request_has_settings(self, type_):
        """
        Determine if the current request (GET query string) contains any
        filter/sort settings for the grid.
        """
        if type_ == 'filter':
            for filtr in self.iter_filters():
                if filtr.key in self.request.GET:
                    return True
            if 'filter' in self.request.GET: # user may be applying empty filters
                return True

        elif type_ == 'sort':
            for key in ['sortkey', 'sortdir']:
                if key in self.request.GET:
                    return True

        elif type_ == 'page':
            for key in ['pagesize', 'page']:
                if key in self.request.GET:
                    return True

        return False

    def session_has_settings(self):
        """
        Determine if the current session contains any settings for the grid.
        """
        # session should have all or nothing, so just check a few keys which
        # should be guaranteed present if anything has been stashed
        for key in ['page', 'sortkey']:
            if 'grid.{}.{}'.format(self.key, key) in self.request.session:
                return True
        return any([key.startswith('grid.{}.filter'.format(self.key)) for key in self.request.session])

    def get_setting(self, source, settings, key, normalize=lambda v: v, default=None):
        """
        Get the effective value for a particular setting, preferring ``source``
        but falling back to existing ``settings`` and finally the ``default``.
        """
        if source not in ('request', 'session'):
            raise ValueError("Invalid source identifier: {}".format(source))

        # If source is query string, try that first.
        if source == 'request':
            value = self.request.GET.get(key)
            if value is not None:
                try:
                    value = normalize(value)
                except ValueError:
                    pass
                else:
                    return value

        # Or, if source is session, try that first.
        else:
            value = self.request.session.get('grid.{}.{}'.format(self.key, key))
            if value is not None:
                return normalize(value)

        # If source had nothing, try default/existing settings.
        value = settings.get(key)
        if value is not None:
            try:
                value = normalize(value)
            except ValueError:
                pass
            else:
                return value

        # Okay then, default it is.
        return default

    def update_filter_settings(self, settings, source):
        """
        Updates a settings dictionary according to filter settings data found
        in either the GET query string, or session storage.

        :param settings: Dictionary of initial settings, which is to be updated.

        :param source: String identifying the source to consult for settings
           data.  Must be one of: ``('request', 'session')``.
        """
        if not self.filterable:
            return

        for filtr in self.iter_filters():
            prefix = 'filter.{}'.format(filtr.key)

            if source == 'request':
                # consider filter active if query string contains a value for it
                settings['{}.active'.format(prefix)] = filtr.key in self.request.GET
                settings['{}.verb'.format(prefix)] = self.get_setting(
                    source, settings, '{}.verb'.format(filtr.key), default='')
                settings['{}.value'.format(prefix)] = self.get_setting(
                    source, settings, filtr.key, default='')

            else: # source = session
                settings['{}.active'.format(prefix)] = self.get_setting(
                    source, settings, '{}.active'.format(prefix),
                    normalize=lambda v: six.text_type(v).lower() == 'true', default=False)
                settings['{}.verb'.format(prefix)] = self.get_setting(
                    source, settings, '{}.verb'.format(prefix), default='')
                settings['{}.value'.format(prefix)] = self.get_setting(
                    source, settings, '{}.value'.format(prefix), default='')

    def update_sort_settings(self, settings, source):
        """
        Updates a settings dictionary according to sort settings data found in
        either the GET query string, or session storage.

        :param settings: Dictionary of initial settings, which is to be updated.

        :param source: String identifying the source to consult for settings
           data.  Must be one of: ``('request', 'session')``.
        """
        if not self.sortable:
            return
        settings['sortkey'] = self.get_setting(source, settings, 'sortkey')
        settings['sortdir'] = self.get_setting(source, settings, 'sortdir')

    def update_page_settings(self, settings):
        """
        Updates a settings dictionary according to pager settings data found in
        either the GET query string, or session storage.

        Note that due to how the actual pager functions, the effective settings
        will often come from *both* the request and session.  This is so that
        e.g. the page size will remain constant (coming from the session) while
        the user jumps between pages (which only provides the single setting).

        :param settings: Dictionary of initial settings, which is to be updated.
        """
        if not self.pageable:
            return

        pagesize = self.request.GET.get('pagesize')
        if pagesize is not None:
            if pagesize.isdigit():
                settings['pagesize'] = int(pagesize)
        else:
            pagesize = self.request.session.get('grid.{}.pagesize'.format(self.key))
            if pagesize is not None:
                settings['pagesize'] = pagesize

        page = self.request.GET.get('page')
        if page is not None:
            if page.isdigit():
                settings['page'] = int(page)
        else:
            page = self.request.session.get('grid.{}.page'.format(self.key))
            if page is not None:
                settings['page'] = int(page)

    def persist_settings(self, settings, to='session'):
        """
        Persist the given settings in some way, as defined by ``func``.
        """
        def persist(key, value=lambda k: settings[k]):
            if to == 'defaults':
                skey = 'tailbone.{}.grid.{}.{}'.format(self.request.user.uuid, self.key, key)
                api.save_setting(Session(), skey, value(key))
            else: # to == session
                skey = 'grid.{}.{}'.format(self.key, key)
                self.request.session[skey] = value(key)

        if self.filterable:
            for filtr in self.iter_filters():
                persist('filter.{}.active'.format(filtr.key), value=lambda k: six.text_type(settings[k]).lower())
                persist('filter.{}.verb'.format(filtr.key))
                persist('filter.{}.value'.format(filtr.key))

        if self.sortable:
            persist('sortkey')
            persist('sortdir')

        if self.pageable:
            persist('pagesize')
            persist('page')

    def filter_data(self, data):
        """
        Filter and return the given data set, according to current settings.
        """
        for filtr in self.iter_active_filters():

            # apply filter to data but save reference to original; if data is a
            # SQLAlchemy query and wasn't modified, we don't need to bother
            # with the underlying join (if there is one)
            original = data
            data = filtr.filter(data)
            if filtr.key in self.joiners and filtr.key not in self.joined and (
                    not isinstance(data, orm.Query) or data is not original):

                # this filter requires a join; apply that
                data = self.joiners[filtr.key](data)
                self.joined.add(filtr.key)

        return data

    def sort_data(self, data):
        """
        Sort the given query according to current settings, and return the result.
        """
        # Cannot sort unless we know which column to sort by.
        if not self.sortkey:
            return data

        # Cannot sort unless we have a sort function.
        sortfunc = self.sorters.get(self.sortkey)
        if not sortfunc:
            return data

        # We can provide a default sort direction though.
        sortdir = getattr(self, 'sortdir', 'asc')
        if self.sortkey in self.joiners and self.sortkey not in self.joined:
            data = self.joiners[self.sortkey](data)
            self.joined.add(self.sortkey)
        return sortfunc(data, sortdir)

    def paginate_data(self, data):
        """
        Paginate the given data set according to current settings, and return
        the result.
        """
        # we of course assume our current page is correct, at first
        pager = self.make_pager(data)

        # if pager has detected that our current page is outside the valid
        # range, we must re-orient ourself around the "new" (valid) page
        if pager.page != self.page:
            self.page = pager.page
            self.request.session['grid.{}.page'.format(self.key)] = self.page
            pager = self.make_pager(data)

        return pager

    def make_pager(self, data):
        return SqlalchemyOrmPage(data,
                                 items_per_page=self.pagesize,
                                 page=self.page,
                                 url_maker=URLMaker(self.request))

    def make_visible_data(self):
        """
        Apply various settings to the raw data set, to produce a final data
        set.  This will page / sort / filter as necessary, according to the
        grid's defaults and the current request etc.
        """
        self.joined = set()
        data = self.data
        if self.filterable:
            data = self.filter_data(data)
        if self.sortable:
            data = self.sort_data(data)
        if self.pageable:
            self.pager = self.paginate_data(data)
            data = self.pager
        return data

    def render_complete(self, template='/grids/complete.mako', **kwargs):
        """
        Render the complete grid, including filters.
        """
        context = kwargs
        context['grid'] = self
        context['request'] = self.request
        context.setdefault('allow_save_defaults', True)
        return render(template, context)

    def render_buefy(self, template='/grids/buefy.mako', **kwargs):
        """
        Render the Buefy grid, complete with filters.  Note that this also
        includes the context menu items and grid tools.
        """
        if 'grid_columns' not in kwargs:
            kwargs['grid_columns'] = self.get_buefy_columns()

        if 'grid_data' not in kwargs:
            kwargs['grid_data'] = self.get_buefy_data()

        if 'static_data' not in kwargs:
            kwargs['static_data'] = self.has_static_data()

        if self.filterable and 'filters_data' not in kwargs:
            kwargs['filters_data'] = self.get_filters_data()

        if self.filterable and 'filters_sequence' not in kwargs:
            kwargs['filters_sequence'] = self.get_filters_sequence()

        return self.render_complete(template=template, **kwargs)

    def render_buefy_table_element(self, template='/grids/b-table.mako', data_prop='gridData', **kwargs):
        """
        This is intended for ad-hoc "small" grids with static data.  Renders
        just a ``<b-table>`` element instead of the typical "full" grid.
        """
        context = dict(kwargs)
        context['grid'] = self
        context['data_prop'] = data_prop
        if 'grid_columns' not in context:
            context['grid_columns'] = self.get_buefy_columns()
        return render(template, context)

    def get_filters_sequence(self):
        """
        Returns a list of filter keys (strings) in the sequence with which they
        should be displayed in the UI.
        """
        return list(self.filters)

    def get_filters_data(self):
        """
        Returns a dict of current filters data, for use with Buefy grid view.
        """
        data = {}
        for filtr in self.filters.values():

            valueless = [v for v in filtr.valueless_verbs
                         if v in filtr.verbs]

            choices = []
            choice_labels = {}
            if self.enums and filtr.key in self.enums:
                choices = list(self.enums[filtr.key])
                choice_labels = self.enums[filtr.key]

            data[filtr.key] = {
                'key': filtr.key,
                'label': filtr.label,
                'active': filtr.active,
                'visible': filtr.active,
                'verbs': filtr.verbs,
                'valueless_verbs': valueless,
                'verb_labels': filtr.verb_labels,
                'verb': filtr.verb or filtr.default_verb or filtr.verbs[0],
                'value': six.text_type(filtr.value) if filtr.value is not None else "",
                'data_type': filtr.data_type,
                'choices': choices,
                'choice_labels': choice_labels,
            }

        return data

    def render_filters(self, template='/grids/filters.mako', **kwargs):
        """
        Render the filters to a Unicode string, using the specified template.
        Additional kwargs are passed along as context to the template.
        """
        # Provide default data to filters form, so renderer can do some of the
        # work for us.
        data = {}
        for filtr in self.iter_active_filters():
            data['{}.active'.format(filtr.key)] = filtr.active
            data['{}.verb'.format(filtr.key)] = filtr.verb
            data[filtr.key] = filtr.value

        form = gridfilters.GridFiltersForm(self.filters,
                                           request=self.request,
                                           defaults=data)

        kwargs['request'] = self.request
        kwargs['grid'] = self
        kwargs['form'] = form
        return render(template, kwargs)

    def render_actions(self, row, i):
        """
        Returns the rendered contents of the 'actions' column for a given row.
        """
        main_actions = [self.render_action(a, row, i)
                        for a in self.main_actions]
        main_actions = [a for a in main_actions if a]
        more_actions = [self.render_action(a, row, i)
                        for a in self.more_actions]
        more_actions = [a for a in more_actions if a]
        if more_actions:
            icon = HTML.tag('span', class_='ui-icon ui-icon-carat-1-e')
            link = tags.link_to("More" + icon, '#', class_='more')
            main_actions.append(HTML.literal('&nbsp; ') + link + HTML.tag('div', class_='more', c=more_actions))
        return HTML.literal('').join(main_actions)

    def render_action(self, action, row, i):
        """
        Renders an action menu item (link) for the given row.
        """
        url = action.get_url(row, i)
        if url:
            kwargs = {'class_': action.key, 'target': action.target}
            if action.icon:
                icon = HTML.tag('span', class_='ui-icon ui-icon-{}'.format(action.icon))
                return tags.link_to(icon + action.label, url, **kwargs)
            return tags.link_to(action.label, url, **kwargs)

    def get_row_key(self, item):
        """
        Must return a unique key for the given data item's row.
        """
        mapper = orm.object_mapper(item)
        if len(mapper.primary_key) == 1:
            return getattr(item, mapper.primary_key[0].key)
        raise NotImplementedError

    def checkbox(self, item):
        """
        Returns boolean indicating whether a checkbox should be rendererd for
        the given data item's row.
        """
        return True

    def render_checkbox(self, item):
        """
        Renders a checkbox cell for the given item, if applicable.
        """
        if not self.checkbox(item):
            return ''
        return tags.checkbox('checkbox-{}-{}'.format(self.key, self.get_row_key(item)),
                             checked=self.checked(item))

    def get_pagesize_options(self):

        # use values from config, if defined
        options = self.request.rattail_config.getlist('tailbone', 'grid.pagesize_options')
        if options:
            options = [int(size) for size in options
                       if size.isdigit()]
            if options:
                return options

        return [5, 10, 20, 50, 100, 200]

    def has_static_data(self):
        """
        Should return ``True`` if the grid data can be considered "static"
        (i.e. a list of values).  Will return ``False`` otherwise, e.g. if the
        data is represented as a SQLAlchemy query.
        """
        # TODO: should make this smarter?
        if isinstance(self.data, list):
            return True
        return False

    def get_buefy_columns(self):
        """
        Return a list of dicts representing all grid columns.  Meant for use
        with Buefy table.
        """
        columns = []
        for name in self.columns:
            columns.append({
                'field': name,
                'label': self.get_label(name),
                'sortable': self.sortable and name in self.sorters,
            })
        return columns

    def get_buefy_data(self):
        """
        Returns a list of data rows for the grid, for use with Buefy table.
        """
        # filter / sort / paginate to get "visible" data
        raw_data = self.make_visible_data()
        data = []
        status_map = {}

        # iterate over data rows
        for i in range(len(raw_data)):
            rowobj = raw_data[i]
            row = {}

            # sometimes we need to include some "raw" data columns in our
            # result set, even though the column is not displayed as part of
            # the grid.  this can be used for front-end editing of row data for
            # instance, when the "display" version is different than raw data.
            # here is the hack we use for that.
            columns = list(self.columns)
            if hasattr(self, 'buefy_data_columns'):
                columns.extend(self.buefy_data_columns)

            # iterate over data fields
            for name in columns:

                # leverage configured rendering logic where applicable;
                # otherwise use "raw" data value as string
                if self.renderers and name in self.renderers:
                    value = self.renderers[name](rowobj, name)
                else:
                    value = self.obtain_value(rowobj, name)
                if value is None:
                    value = ""
                row[name] = six.text_type(value)

            # maybe add UUID for convenience
            if 'uuid' not in self.columns:
                if hasattr(rowobj, 'uuid'):
                    row['uuid'] = rowobj.uuid

            # set action URL(s) for row, as needed
            self.set_action_urls(row, rowobj, i)

            # set extra row class if applicable
            if self.extra_row_class:
                status = self.extra_row_class(rowobj, i)
                if status:
                    status_map[i] = status

            data.append(row)

        results = {
            'data': data,
            'row_status_map': status_map,
        }

        if self.pageable and self.pager is not None:
            results['total_items'] = self.pager.item_count
            results['per_page'] = self.pager.items_per_page
            results['page'] = self.pager.page
            results['pages'] = self.pager.page_count
            results['first_item'] = self.pager.first_item
            results['last_item'] = self.pager.last_item

        return results

    def set_action_urls(self, row, rowobj, i):
        """
        Pre-generate all action URLs for the given data row.  Meant for use
        with Buefy table, since we can't generate URLs from JS.
        """
        for action in (self.main_actions + self.more_actions):
            url = action.get_url(rowobj, i)
            row['_action_url_{}'.format(action.key)] = url

    def is_linked(self, name):
        """
        Should return ``True`` if the given column name is configured to be
        "linked" (i.e. table cell should contain a link to "view object"),
        otherwise ``False``.
        """
        if self.linked_columns:
            if name in self.linked_columns:
                return True
        return False


class CustomWebhelpersGrid(webhelpers2_grid.Grid):
    """
    Implement column sorting links etc. for webhelpers2_grid
    """

    def __init__(self, itemlist, columns, **kwargs):
        self.mobile = kwargs.pop('mobile', False)
        self.renderers = kwargs.pop('renderers', {})
        self.linked_columns = kwargs.pop('linked_columns', [])
        self.extra_record_class = kwargs.pop('extra_record_class', None)
        super(CustomWebhelpersGrid, self).__init__(itemlist, columns, **kwargs)

    def default_header_record_format(self, headers):
        if self.mobile:
            return HTML('')
        return super(CustomWebhelpersGrid, self).default_header_record_format(headers)

    def generate_header_link(self, column_number, column, label_text):

        # display column header as simple no-op link; client-side JS takes care
        # of the rest for us
        label_text = tags.link_to(label_text, '#', data_sortkey=column)

        # Is the current column the one we're ordering on?
        if (column == self.order_column):
            return self.default_header_ordered_column_format(column_number,
                                                             column,
                                                             label_text)
        else:
            return self.default_header_column_format(column_number, column,
                                                     label_text)            

    def default_record_format(self, i, record, columns):
        if self.mobile:
            return columns
        kwargs = {
            'class_': self.get_record_class(i, record, columns),
        }
        if hasattr(record, 'uuid'):
            kwargs['data_uuid'] = record.uuid
        return HTML.tag('tr', columns, **kwargs)

    def get_record_class(self, i, record, columns):
        if i % 2 == 0:
            cls = 'even r{}'.format(i)
        else:
            cls = 'odd r{}'.format(i)
        if self.extra_record_class:
            extra = self.extra_record_class(record, i)
            if extra:
                cls = '{} {}'.format(cls, extra)
        return cls

    def get_column_value(self, column_number, i, record, column_name):
        if self.renderers and column_name in self.renderers:
            return self.renderers[column_name](record, column_name)
        try:
            return record[column_name]
        except TypeError:
            return getattr(record, column_name)

    def default_column_format(self, column_number, i, record, column_name):
        value = self.get_column_value(column_number, i, record, column_name)
        if self.mobile:
            url = self.url_generator(record, i)
            attrs = {}
            if hasattr(record, 'uuid'):
                attrs['data_uuid'] = record.uuid
            return HTML.tag('li', tags.link_to(value, url), **attrs)
        if self.linked_columns and column_name in self.linked_columns and (
                value is not None and value != ''):
            url = self.url_generator(record, i)
            value = tags.link_to(value, url)
        class_name = 'c{} {}'.format(column_number, column_name)
        return HTML.tag('td', value, class_=class_name)


class GridAction(object):
    """
    Represents an action available to a grid.  This is used to construct the
    'actions' column when rendering the grid.
    """

    def __init__(self, key, label=None, url='#', icon=None, target=None,
                 click_handler=None):
        self.key = key
        self.label = label or prettify(key)
        self.icon = icon
        self.url = url
        self.target = target
        self.click_handler = click_handler

    def get_url(self, row, i):
        """
        Returns an action URL for the given row.
        """
        if callable(self.url):
            return self.url(row, i)
        return self.url


class URLMaker(object):
    """
    URL constructor for use with SQLAlchemy grid pagers.  Logic for this was
    basically copied from the old `webhelpers.paginate` module
    """

    def __init__(self, request):
        self.request = request

    def __call__(self, page):
        params = self.request.GET.copy()
        params["page"] = page
        params["partial"] = "1"
        qs = urllib.parse.urlencode(params, True)
        return '{}?{}'.format(self.request.path, qs)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for inventory batches
"""

from __future__ import unicode_literals, absolute_import

import re
import decimal
import logging

import six

from rattail import pod
from rattail.db import model, api
from rattail.db.util import make_full_description
from rattail.time import localtime
from rattail.gpc import GPC
from rattail.util import pretty_quantity

import colander
from deform import widget as dfwidget
from webhelpers2.html import HTML, tags

from tailbone import forms, grids
from tailbone.views import MasterView
from tailbone.views.batch import BatchMasterView


log = logging.getLogger(__name__)


class InventoryAdjustmentReasonsView(MasterView):
    """
    Master view for inventory adjustment reasons.
    """
    model_class = model.InventoryAdjustmentReason
    route_prefix = 'invadjust_reasons'
    url_prefix = '/inventory-adjustment-reasons'

    grid_columns = [
        'code',
        'description',
        'hidden',
    ]

    def configure_grid(self, g):
        super(InventoryAdjustmentReasonsView, self).configure_grid(g)
        g.set_sort_defaults('code')

    def configure_form(self, f):
        super(InventoryAdjustmentReasonsView, self).configure_form(f)

        # code
        f.set_validator('code', self.unique_code)

    def unique_code(self, node, value):
        query = self.Session.query(model.InventoryAdjustmentReason)\
                            .filter(model.InventoryAdjustmentReason.code == value)
        if self.editing:
            reason = self.get_instance()
            query = query.filter(model.InventoryAdjustmentReason.uuid != reason.uuid)
        if query.count():
            raise colander.Invalid(node, "Code must be unique")


class InventoryBatchView(BatchMasterView):
    """
    Master view for inventory batches.
    """
    model_class = model.InventoryBatch
    model_title_plural = "Inventory Batches"
    default_handler_spec = 'rattail.batch.inventory:InventoryBatchHandler'
    route_prefix = 'batch.inventory'
    url_prefix = '/batch/inventory'
    index_title = "Inventory"
    rows_creatable = True
    results_executable = True
    mobile_creatable = True
    mobile_rows_creatable = True

    # set to False to disable "zero all" batch count mode
    allow_zero_all = True

    # set to False to disable "variance" batch count mode
    allow_variance = True

    # set to False to prevent exposing case fields for user input,
    # when the batch count mode is "adjust only"
    allow_adjustment_cases = True

    # set to True for the UI to "prefer" case amounts, as opposed to unit
    prefer_cases = False

    labels = {
        'mode': "Count Mode",
    }

    grid_columns = [
        'id',
        'created',
        'created_by',
        'description',
        'mode',
        'rowcount',
        'total_cost',
        'executed',
        'executed_by',
    ]

    form_fields = [
        'id',
        'description',
        'notes',
        'created',
        'created_by',
        'handheld_batches',
        'mode',
        'reason_code',
        'total_cost',
        'rowcount',
        'complete',
        'executed',
        'executed_by',
    ]

    mobile_form_fields = [
        'mode',
        'reason_code',
        'rowcount',
        'complete',
        'executed',
        'executed_by',
    ]

    model_row_class = model.InventoryBatchRow
    rows_editable = True

    row_labels = {
        'upc': "UPC",
        'previous_units_on_hand': "Prev. On Hand",
    }

    row_grid_columns = [
        'sequence',
        'upc',
        'item_id',
        'brand_name',
        'description',
        'size',
        'previous_units_on_hand',
        'cases',
        'units',
        'unit_cost',
        'total_cost',
        'status_code',
    ]

    row_form_fields = [
        'sequence',
        'upc',
        'brand_name',
        'description',
        'size',
        'status_code',
        'previous_units_on_hand',
        'case_quantity',
        'cases',
        'units',
        'unit_cost',
        'total_cost',
        'variance',
    ]

    def configure_grid(self, g):
        super(InventoryBatchView, self).configure_grid(g)

        # mode
        g.set_enum('mode', self.enum.INVENTORY_MODE)
        g.filters['mode'].set_value_renderer(
            grids.filters.EnumValueRenderer(self.enum.INVENTORY_MODE))

        # total_cost
        g.set_type('total_cost', 'currency')

    def render_mobile_listitem(self, batch, i):
        return "({}) {} rows - {}, {}".format(
            batch.id_str,
            "?" if batch.rowcount is None else batch.rowcount,
            batch.created_by,
            localtime(self.request.rattail_config, batch.created, from_utc=True).strftime('%Y-%m-%d'))

    def mutable_batch(self, batch):
        return not batch.executed and not batch.complete and batch.mode != self.enum.INVENTORY_MODE_ZERO_ALL

    def allow_worksheet(self, batch):
        return self.mutable_batch(batch)

    def get_available_modes(self):
        permission_prefix = self.get_permission_prefix()
        modes = dict(self.enum.INVENTORY_MODE)
        if not self.request.has_perm('{}.create.replace'.format(permission_prefix)):
            if hasattr(self.enum, 'INVENTORY_MODE_REPLACE'):
                modes.pop(self.enum.INVENTORY_MODE_REPLACE, None)
            if hasattr(self.enum, 'INVENTORY_MODE_REPLACE_ADJUST'):
                modes.pop(self.enum.INVENTORY_MODE_REPLACE_ADJUST, None)
        if not self.allow_zero_all or not self.request.has_perm('{}.create.zero'.format(permission_prefix)):
            if hasattr(self.enum, 'INVENTORY_MODE_ZERO_ALL'):
                modes.pop(self.enum.INVENTORY_MODE_ZERO_ALL, None)
        if not self.allow_variance or not self.request.has_perm('{}.create.variance'.format(permission_prefix)):
            if hasattr(self.enum, 'INVENTORY_MODE_VARIANCE'):
                modes.pop(self.enum.INVENTORY_MODE_VARIANCE, None)
        return modes

    def configure_form(self, f):
        super(InventoryBatchView, self).configure_form(f)

        # mode
        modes = self.get_available_modes()
        f.set_enum('mode', modes)
        f.set_label('mode', "Count Mode")
        if len(modes) == 1:
            f.set_widget('mode', forms.widgets.ReadonlyWidget())
            f.set_default('mode', list(modes)[0])

        # total_cost
        if self.creating:
            f.remove_field('total_cost')
        else:
            f.set_readonly('total_cost')
            f.set_type('total_cost', 'currency')

        # handheld_batches
        if self.creating:
            f.remove_field('handheld_batches')
        else:
            f.set_readonly('handheld_batches')
            f.set_renderer('handheld_batches', self.render_handheld_batches)

        # complete
        if self.creating:
            f.remove_field('complete')

    def render_handheld_batches(self, inventory_batch, field):
        items = ''
        for handheld in inventory_batch._handhelds:
            text = handheld.handheld.id_str
            url = self.request.route_url('batch.handheld.view', uuid=handheld.handheld_uuid)
            items += HTML.tag('li', c=tags.link_to(text, url))
        return HTML.tag('ul', c=items)

    def row_editable(self, row):
        return self.mutable_batch(row.batch)

    def row_deletable(self, row):
        return self.mutable_batch(row.batch)

    def save_edit_row_form(self, form):
        row = form.model_instance
        batch = row.batch
        if batch.total_cost is not None and row.total_cost is not None:
            batch.total_cost -= row.total_cost
        return super(InventoryBatchView, self).save_edit_row_form(form)

    def delete_row(self):
        row = self.Session.query(model.InventoryBatchRow).get(self.request.matchdict['row_uuid'])
        if not row:
            raise self.notfound()
        batch = row.batch
        if batch.total_cost is not None and row.total_cost is not None:
            batch.total_cost -= row.total_cost
        return super(InventoryBatchView, self).delete_row()

    def create_row(self):
        """
        Desktop workflow view for adding items to inventory batch.
        """
        batch = self.get_instance()
        if batch.executed:
            return self.redirect(self.get_action_url('view', batch))

        schema = DesktopForm().bind(session=self.Session())
        form = forms.Form(schema=schema, request=self.request)
        if form.validate(newstyle=True):

            product = self.Session.query(model.Product).get(form.validated['product'])

            row = None
            if self.should_aggregate_products(batch):
                row = self.find_row_for_product(batch, product)
                if row:
                    row.cases = form.validated['cases']
                    row.units = form.validated['units']
                    self.handler.refresh_row(row)

            if not row:
                row = model.InventoryBatchRow()
                row.product = product
                row.upc = form.validated['upc']
                row.brand_name = form.validated['brand_name']
                row.description = form.validated['description']
                row.size = form.validated['size']
                row.case_quantity = form.validated['case_quantity']
                row.cases = form.validated['cases']
                row.units = form.validated['units']
                self.handler.capture_current_units(row)
                self.handler.add_row(batch, row)

            description = make_full_description(form.validated['brand_name'],
                                                form.validated['description'],
                                                form.validated['size'])
            self.request.session.flash("{} cases, {} units: {} {}".format(
                form.validated['cases'] or 0, form.validated['units'] or 0,
                form.validated['upc'].pretty(), description))
            return self.redirect(self.request.current_route_url())

        title = self.get_instance_title(batch)
        return self.render_to_response('desktop_form', {
            'batch': batch,
            'instance': batch,
            'instance_title': title,
            'index_title': "{}: {}".format(self.get_model_title(), title),
            'index_url': self.get_action_url('view', batch),
            'form': form,
            'dform': form.make_deform_form(),
            'allow_cases': self.allow_cases(batch),
            'prefer_cases': self.prefer_cases,
        })

    def allow_cases(self, batch):
        if batch.mode == self.enum.INVENTORY_MODE_ADJUST:
            if self.allow_adjustment_cases:
                return True
            return False
        return True

    def should_aggregate_products(self, batch):
        """
        Must return a boolean indicating whether rows should be aggregated by
        product for the given batch.
        """
        if batch.mode == self.enum.INVENTORY_MODE_VARIANCE:
            return True
        return False

    def desktop_lookup(self):
        """
        Try to locate a product by UPC, and validate it in the context of
        current batch, returning some data for client JS.
        """
        batch = self.get_instance()
        if batch.executed:
            return {
                'error': "Current batch has already been executed",
                'redirect': self.get_action_url('view', batch),
            }
        entry = self.request.GET.get('upc', '')
        aggregate = self.should_aggregate_products(batch)

        type2 = self.find_type2_product(entry)
        if type2:
            product, price = type2
        else:
            product = self.find_product(entry)

        force_unit_item = True # TODO: make configurable?
        unit_forced = False
        if force_unit_item and product and product.is_pack_item():
            product = product.unit
            unit_forced = True

        data = self.product_info(product)
        if type2:
            data['type2'] = True
            if not aggregate:
                if price is None:
                    data['units'] = 1
                else:
                    data['units'] = float((price / product.regular_price.price).quantize(decimal.Decimal('0.01')))

        result = {'product': data, 'upc_raw': entry, 'upc': None, 'force_unit_item': unit_forced}
        if not data:
            upc = re.sub(r'\D', '', entry.strip())
            if upc:
                upc = GPC(upc)
                result['upc'] = six.text_type(upc)
                result['upc_pretty'] = upc.pretty()
                result['image_url'] = pod.get_image_url(self.rattail_config, upc)

        if product and aggregate:
            row = self.find_row_for_product(batch, product)
            if row:
                result['already_present_in_batch'] = True
                result['cases'] = float(row.cases) if row.cases is not None else None
                result['units'] = float(row.units) if row.units is not None else None

        return result

    def find_row_for_product(self, batch, product):
        rows = self.Session.query(model.InventoryBatchRow)\
                           .filter(model.InventoryBatchRow.batch == batch)\
                           .filter(model.InventoryBatchRow.product == product)\
                           .filter(model.InventoryBatchRow.removed == False)\
                           .all()
        if rows:
            if len(rows) > 1:
                log.error("inventory batch %s should aggregate products, but has %s rows for: %s",
                          batch.id_str, len(rows), product)
            return rows[0]

    def find_product(self, entry):
        upc = re.sub(r'\D', '', entry.strip())
        if upc:

            # first try to locate existing batch row by UPC match
            provided = GPC(upc, calc_check_digit=False)
            checked = GPC(upc, calc_check_digit='upc')
            product = api.get_product_by_upc(self.Session(), provided)
            if product:
                return product
            product = api.get_product_by_upc(self.Session(), checked)
            if product:
                return product

        # maybe try to locate product by alternate code
        if self.rattail_config.getbool('tailbone', 'inventory.lookup_by_code', default=False):
            product = api.get_product_by_code(self.Session(), entry)
            if product:
                return product

    def product_info(self, product):
        data = {}
        if product and (not product.deleted or self.request.has_perm('products.view_deleted')):
            data['uuid'] = product.uuid
            data['upc'] = six.text_type(product.upc)
            data['upc_pretty'] = product.upc.pretty()
            data['full_description'] = product.full_description
            data['brand_name'] = six.text_type(product.brand or '')
            data['description'] = product.description
            data['size'] = product.size
            data['case_quantity'] = 1 # default
            data['cost_found'] = False
            data['image_url'] = pod.get_image_url(self.rattail_config, product.upc)
        return data

    def configure_mobile_form(self, f):
        super(InventoryBatchView, self).configure_mobile_form(f)
        batch = f.model_instance

        # mode
        modes = self.get_available_modes()
        f.set_enum('mode', modes)
        mode_values = [(k, v) for k, v in sorted(modes.items())]
        f.set_widget('mode', forms.widgets.PlainSelectWidget(values=mode_values))

        # complete
        if self.creating or batch.executed or not batch.complete:
            f.remove_field('complete')

        # rowcount
        if self.viewing and not batch.executed and not batch.complete:
            f.remove_field('rowcount')

    # TODO: document this, maybe move it etc.
    unknown_product_creates_row = True

    # TODO: this view can create new rows, with only a GET query.  that should
    # probably be changed to require POST; for now we just require the "create
    # batch row" perm and call it good..
    def mobile_row_from_upc(self):
        """
        Locate and/or create a row within the batch, according to the given
        product UPC, then redirect to the row view page.
        """
        batch = self.get_instance()
        row = None
        raw_entry = self.request.GET.get('upc', '')
        entry = raw_entry.strip()
        entry = re.sub(r'\D', '', entry)
        if entry:

            if len(entry) <= 14:
                row = self.add_row_for_upc(batch, entry, warn_if_present=True)
                if not row:
                    self.request.session.flash("Product not found: {}".format(entry), 'error')
                    return self.redirect(self.get_action_url('view', batch, mobile=True))

            else:
                self.request.session.flash("UPC has too many digits ({}): {}".format(len(entry), entry), 'error')
                return self.redirect(self.get_action_url('view', batch, mobile=True))

        else:
            self.request.session.flash("Product not found: {}".format(raw_entry), 'error')
            return self.redirect(self.get_action_url('view', batch, mobile=True))

        self.Session.flush()
        return self.redirect(self.mobile_row_route_url('view', uuid=row.batch_uuid, row_uuid=row.uuid))

    def add_row_for_upc(self, batch, entry, warn_if_present=False):
        """
        Add a row to the batch for the given UPC, if applicable.
        """
        type2 = self.find_type2_product(entry)
        if type2:
            product, price = type2
        else:
            product = self.find_product(entry)
        if product:

            force_unit_item = True # TODO: make configurable?
            if force_unit_item and product.is_pack_item():
                product = product.unit
                self.request.session.flash("You scanned a pack item, but must count the units instead.", 'error')

            aggregate = self.should_aggregate_products(batch)
            if aggregate:
                row = self.find_row_for_product(batch, product)
                if row:
                    if warn_if_present:
                        self.request.session.flash("Product already exists in batch; please confirm counts", 'error')
                    return row

            row = model.InventoryBatchRow()
            row.product = product
            row.upc = product.upc
            self.handler.capture_current_units(row)
            if type2 and not aggregate:
                if price is None:
                    row.units = 1
                else:
                    row.units = (price / product.regular_price.price).quantize(decimal.Decimal('0.01'))
            self.handler.add_row(batch, row)
            return row

        elif self.unknown_product_creates_row:
            row = model.InventoryBatchRow()
            row.upc = GPC(upc, calc_check_digit=False) # TODO: why not calc check digit?
            row.description = "(unknown product)"
            self.handler.capture_current_units(row)
            self.handler.add_row(batch, row)
            return row

    def template_kwargs_view_row(self, **kwargs):
        row = kwargs['instance']
        kwargs['product_image_url'] = pod.get_image_url(self.rattail_config, row.upc)
        return kwargs

    def get_batch_kwargs(self, batch, mobile=False):
        kwargs = super(InventoryBatchView, self).get_batch_kwargs(batch, mobile=False)
        kwargs['mode'] = batch.mode
        kwargs['complete'] = False
        kwargs['reason_code'] = batch.reason_code
        return kwargs

    def get_mobile_row_data(self, batch):
        # we want newest on top, for inventory batch rows
        return self.get_row_data(batch)\
                   .order_by(self.model_row_class.sequence.desc())

    # TODO: ugh, the hackiness.  needs a refactor fo sho
    def mobile_view_row(self):
        """
        Mobile view for inventory batch rows.  Note that this also handles
        updating a row...ugh.
        """
        self.viewing = True
        row = self.get_row_instance()
        batch = self.get_parent(row)
        form = self.make_mobile_row_form(row)

        allow_cases = self.allow_cases(batch)
        unit_uom = 'LB' if row.product and row.product.weighed else 'EA'
        if row.cases and allow_cases:
            uom = 'CS'
        elif row.units:
            uom = unit_uom
        elif row.case_quantity and allow_cases and self.prefer_cases:
            uom = 'CS'
        else:
            uom = unit_uom

        context = {
            'row': row,
            'batch': batch,
            'instance': row,
            'instance_title': self.get_row_instance_title(row),
            'parent_model_title': self.get_model_title(),
            'parent_title': self.get_instance_title(batch),
            'parent_url': self.get_action_url('view', batch, mobile=True),
            'product_image_url': pod.get_image_url(self.rattail_config, row.upc),
            'form': form,
            'allow_cases': allow_cases,
            'unit_uom': unit_uom,
            'uom': uom,
        }

        if self.request.has_perm('{}.edit_row'.format(self.get_permission_prefix())):
            schema = InventoryForm().bind(session=self.Session())
            update_form = forms.Form(schema=schema, request=self.request)
            if update_form.validate(newstyle=True):
                row = self.Session.query(model.InventoryBatchRow).get(update_form.validated['row'])
                cases = update_form.validated['cases']
                units = update_form.validated['units']
                if cases is not colander.null:
                    row.cases = cases
                    row.units = None
                elif units is not colander.null:
                    row.cases = None
                    row.units = units
                else:
                    raise NotImplementedError
                self.handler.refresh_row(row)
                route_prefix = self.get_route_prefix()
                return self.redirect(self.request.route_url('mobile.{}.view'.format(route_prefix), uuid=batch.uuid))

        return self.render_to_response('view_row', context, mobile=True)

    def get_row_instance_title(self, row):
        if row.upc:
            return row.upc.pretty()
        if row.item_id:
            return row.item_id
        return "row {}".format(row.sequence)

    def configure_row_grid(self, g):
        super(InventoryBatchView, self).configure_row_grid(g)

        # quantity fields
        g.set_type('previous_units_on_hand', 'quantity')
        g.set_type('cases', 'quantity')
        g.set_type('units', 'quantity')

        # currency fields
        g.set_type('unit_cost', 'currency')
        g.set_type('total_cost', 'currency')

        # short labels
        g.set_label('brand_name', "Brand")
        g.set_label('status_code', "Status")

        # links
        g.set_link('upc')
        g.set_link('item_id')
        g.set_link('description')

    def row_grid_extra_class(self, row, i):
        if row.status_code == row.STATUS_PRODUCT_NOT_FOUND:
            return 'warning'

    def render_mobile_row_listitem(self, row, i):
        description = row.product.full_description if row.product else row.description
        unit_uom = 'LB' if row.product and row.product.weighed else 'EA'
        qty = "{} {}".format(pretty_quantity(row.cases or row.units), 'CS' if row.cases else unit_uom)
        return "({}) {} - {}".format(row.upc.pretty(), description, qty)

    def configure_row_form(self, f):
        super(InventoryBatchView, self).configure_row_form(f)
        row = f.model_instance

        # readonly fields
        f.set_readonly('upc')
        f.set_readonly('item_id')
        f.set_readonly('brand_name')
        f.set_readonly('description')
        f.set_readonly('size')
        f.set_readonly('previous_units_on_hand')
        f.set_readonly('case_quantity')
        f.set_readonly('variance')
        f.set_readonly('total_cost')

        # quantity fields
        f.set_type('case_quantity', 'quantity')
        f.set_type('previous_units_on_hand', 'quantity')
        f.set_type('cases', 'quantity')
        f.set_type('units', 'quantity')
        f.set_type('variance', 'quantity')

        # currency fields
        f.set_type('unit_cost', 'currency')
        f.set_type('total_cost', 'currency')

        # upc
        f.set_renderer('upc', self.render_upc)

        # cases
        if self.editing:
            if not self.allow_cases(row.batch):
                f.set_readonly('cases')

    def render_upc(self, row, field):
        upc = row.upc
        if not upc:
            return ""
        text = upc.pretty()
        if row.product_uuid:
            url = self.request.route_url('products.view', uuid=row.product_uuid)
            return tags.link_to(text, url)
        return text

    @classmethod
    def defaults(cls, config):
        cls._batch_defaults(config)
        cls._defaults(config)
        cls._inventory_defaults(config)

    @classmethod
    def _inventory_defaults(cls, config):
        model_key = cls.get_model_key()
        model_title = cls.get_model_title()
        route_prefix = cls.get_route_prefix()
        url_prefix = cls.get_url_prefix()
        permission_prefix = cls.get_permission_prefix()

        # extra perms for creating batches per "mode"
        config.add_tailbone_permission(permission_prefix, '{}.create.replace'.format(permission_prefix),
                                       "Create new {} with 'replace' mode".format(model_title))
        if cls.allow_zero_all:
            config.add_tailbone_permission(permission_prefix, '{}.create.zero'.format(permission_prefix),
                                           "Create new {} with 'zero' mode".format(model_title))
        if cls.allow_variance:
            config.add_tailbone_permission(permission_prefix, '{}.create.variance'.format(permission_prefix),
                                           "Create new {} with 'variance' mode".format(model_title))

        # row UPC lookup, for desktop
        config.add_route('{}.desktop_lookup'.format(route_prefix), '{}/{{{}}}/desktop-form/lookup'.format(url_prefix, model_key))
        config.add_view(cls, attr='desktop_lookup', route_name='{}.desktop_lookup'.format(route_prefix),
                        renderer='json', permission='{}.create_row'.format(permission_prefix))

        # mobile - make new row from UPC
        config.add_route('mobile.{}.row_from_upc'.format(route_prefix), '/mobile{}/{{{}}}/row-from-upc'.format(url_prefix, model_key))
        config.add_view(cls, attr='mobile_row_from_upc', route_name='mobile.{}.row_from_upc'.format(route_prefix),
                        permission='{}.create_row'.format(permission_prefix))


# TODO: this is a stopgap measure to fix an obvious bug, which exists when the
# session is not provided by the view at runtime (i.e. when it was instead
# being provided by the type instance, which was created upon app startup).
@colander.deferred
def valid_inventory_batch_row(node, kw):
    session = kw['session']
    def validate(node, value):
        row = session.query(model.InventoryBatchRow).get(value)
        if not row:
            raise colander.Invalid(node, "Batch row not found")
        if row.batch.executed:
            raise colander.Invalid(node, "Batch has already been executed")
        return row.uuid
    return validate


class InventoryForm(colander.MappingSchema):

    row = colander.SchemaNode(colander.String(),
                              validator=valid_inventory_batch_row)

    cases = colander.SchemaNode(colander.Decimal(), missing=colander.null)

    units = colander.SchemaNode(colander.Decimal(), missing=colander.null)


# TODO: this is a stopgap measure to fix an obvious bug, which exists when the
# session is not provided by the view at runtime (i.e. when it was instead
# being provided by the type instance, which was created upon app startup).
@colander.deferred
def valid_product(node, kw):
    session = kw['session']
    def validate(node, value):
        product = session.query(model.Product).get(value)
        if not product:
            raise colander.Invalid(node, "Product not found")
        return product.uuid
    return validate


class DesktopForm(colander.Schema):

    product = colander.SchemaNode(colander.String(),
                                  validator=valid_product)

    upc = colander.SchemaNode(forms.types.GPCType())

    brand_name = colander.SchemaNode(colander.String())

    description = colander.SchemaNode(colander.String())

    size = colander.SchemaNode(colander.String(), missing=colander.null)

    case_quantity = colander.SchemaNode(colander.Decimal())

    cases = colander.SchemaNode(colander.Decimal(),
                                missing=None)

    units = colander.SchemaNode(colander.Decimal(),
                                missing=None)


def includeme(config):
    InventoryAdjustmentReasonsView.defaults(config)
    InventoryBatchView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for new product batches
"""

from __future__ import unicode_literals, absolute_import

from rattail.db import model

from tailbone.views.batch import BatchMasterView


class NewProductBatchView(BatchMasterView):
    """
    Master view for new product batches.
    """
    model_class = model.NewProductBatch
    model_row_class = model.NewProductBatchRow
    default_handler_spec = 'rattail.batch.newproduct:NewProductBatchHandler'
    route_prefix = 'batch.newproduct'
    url_prefix = '/batches/newproduct'
    template_prefix = '/batch/newproduct'
    downloadable = True
    rows_editable = True

    form_fields = [
        'id',
        'input_filename',
        'description',
        'notes',
        'created',
        'created_by',
        'rowcount',
        'executed',
        'executed_by',
    ]

    row_labels = {
        'vendor_id': "Vendor ID",
    }

    row_grid_columns = [
        'sequence',
        'upc',
        'brand_name',
        'description',
        'size',
        'vendor',
        'vendor_item_code',
        'department',
        'subdepartment',
        'regular_price',
        'status_code',
    ]

    row_form_fields = [
        'sequence',
        'product',
        'upc',
        'brand_name',
        'description',
        'size',
        'vendor_id',
        'vendor',
        'vendor_item_code',
        'department_number',
        'department',
        'subdepartment_number',
        'subdepartment',
        'case_size',
        'case_cost',
        'unit_cost',
        'regular_price',
        'regular_price_multiple',
        'pack_price',
        'pack_price_multiple',
        'suggested_price',
        'category_code',
        'category',
        'family_code',
        'family',
        'report_code',
        'report',
        'status_code',
        'status_text',
    ]

    def configure_form(self, f):
        super(NewProductBatchView, self).configure_form(f)

        # input_filename
        if self.creating:
            f.set_type('input_filename', 'file')
        else:
            f.set_readonly('input_filename')
            f.set_renderer('input_filename', self.render_downloadable_file)

    def configure_row_grid(self, g):
        super(NewProductBatchView, self).configure_row_grid(g)

        g.set_type('case_cost', 'currency')
        g.set_type('unit_cost', 'currency')
        g.set_type('regular_price', 'currency')
        g.set_type('pack_price', 'currency')
        g.set_type('suggested_price', 'currency')

    def row_grid_extra_class(self, row, i):
        if row.status_code in (row.STATUS_MISSING_KEY,
                               row.STATUS_PRODUCT_EXISTS,
                               row.STATUS_VENDOR_NOT_FOUND,
                               row.STATUS_DEPT_NOT_FOUND,
                               row.STATUS_SUBDEPT_NOT_FOUND):
            return 'warning'
        if row.status_code in (row.STATUS_CATEGORY_NOT_FOUND,
                               row.STATUS_FAMILY_NOT_FOUND,
                               row.STATUS_REPORTCODE_NOT_FOUND):
            return 'notice'

    def configure_row_form(self, f):
        super(NewProductBatchView, self).configure_row_form(f)

        f.set_readonly('product')
        f.set_readonly('vendor')
        f.set_readonly('department')
        f.set_readonly('subdepartment')
        f.set_readonly('category')
        f.set_readonly('family')
        f.set_readonly('report')

        f.set_type('upc', 'gpc')

        f.set_renderer('vendor', self.render_vendor)
        f.set_renderer('department', self.render_department)
        f.set_renderer('subdepartment', self.render_subdepartment)
        f.set_renderer('report', self.render_report)


def includeme(config):
    NewProductBatchView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for handheld batches
"""

from __future__ import unicode_literals, absolute_import

import os

from rattail.db import model
from rattail.util import OrderedDict

import colander
from webhelpers2.html import tags

from tailbone import forms
from tailbone.db import Session
from tailbone.views.batch import FileBatchMasterView


ACTION_OPTIONS = OrderedDict([
    ('make_label_batch', "Make a new Label Batch"),
    ('make_inventory_batch', "Make a new Inventory Batch"),
])


class ExecutionOptions(colander.Schema):

    action = colander.SchemaNode(
        colander.String(),
        validator=colander.OneOf(ACTION_OPTIONS),
        widget=forms.widgets.PlainSelectWidget(values=ACTION_OPTIONS.items()))


class HandheldBatchView(FileBatchMasterView):
    """
    Master view for handheld batches.
    """
    model_class = model.HandheldBatch
    default_handler_spec = 'rattail.batch.handheld:HandheldBatchHandler'
    model_title_plural = "Handheld Batches"
    route_prefix = 'batch.handheld'
    url_prefix = '/batch/handheld'
    execution_options_schema = ExecutionOptions
    editable = False
    results_executable = True

    model_row_class = model.HandheldBatchRow
    rows_creatable = False
    rows_editable = True

    grid_columns = [
        'id',
        'device_type',
        'device_name',
        'created',
        'created_by',
        'rowcount',
        'status_code',
        'executed',
    ]

    form_fields = [
        'id',
        'device_type',
        'device_name',
        'filename',
        'created',
        'created_by',
        'rowcount',
        'status_code',
        'executed',
        'executed_by',
    ]

    row_labels = {
        'upc': "UPC",
    }

    row_grid_columns = [
        'sequence',
        'upc',
        'brand_name',
        'description',
        'size',
        'cases',
        'units',
        'status_code',
    ]

    row_form_fields = [
        'sequence',
        'upc',
        'brand_name',
        'description',
        'size',
        'status_code',
        'cases',
        'units',
    ]

    def configure_grid(self, g):
        super(HandheldBatchView, self).configure_grid(g)
        device_types = OrderedDict(sorted(self.enum.HANDHELD_DEVICE_TYPE.items(),
                                          key=lambda item: item[1]))
        g.set_enum('device_type', device_types)

    def grid_extra_class(self, batch, i):
        if batch.status_code is not None and batch.status_code != batch.STATUS_OK:
            return 'notice'

    def configure_form(self, f):
        super(HandheldBatchView, self).configure_form(f)
        batch = f.model_instance

        # device_type
        device_types = OrderedDict(sorted(self.enum.HANDHELD_DEVICE_TYPE.items(),
                                          key=lambda item: item[1]))
        f.set_enum('device_type', device_types)
        f.widgets['device_type'].values.insert(0, ('', "(none)"))

        if self.creating:
            f.set_fields([
                'filename',
                'device_type',
                'device_name',
            ])

        if self.viewing:
            if batch.inventory_batch:
                f.append('inventory_batch')
                f.set_renderer('inventory_batch', self.render_inventory_batch)

    def render_inventory_batch(self, handheld_batch, field):
        batch = handheld_batch.inventory_batch
        if not batch:
            return ""
        text = batch.id_str
        url = self.request.route_url('batch.inventory.view', uuid=batch.uuid)
        return tags.link_to(text, url)

    def get_batch_kwargs(self, batch):
        kwargs = super(HandheldBatchView, self).get_batch_kwargs(batch)
        kwargs['device_type'] = batch.device_type
        kwargs['device_name'] = batch.device_name
        return kwargs

    def configure_row_grid(self, g):
        super(HandheldBatchView, self).configure_row_grid(g)
        g.set_type('cases', 'quantity')
        g.set_type('units', 'quantity')
        g.set_label('brand_name', "Brand")

    def row_grid_extra_class(self, row, i):
        if row.status_code == row.STATUS_PRODUCT_NOT_FOUND:
            return 'warning'

    def configure_row_form(self, f):
        super(HandheldBatchView, self).configure_row_form(f)

        # readonly fields
        f.set_readonly('upc')
        f.set_readonly('brand_name')
        f.set_readonly('description')
        f.set_readonly('size')

        # upc
        f.set_renderer('upc', self.render_upc)

    def render_upc(self, row, field):
        upc = row.upc
        if not upc:
            return ""
        text = upc.pretty()
        if row.product_uuid:
            url = self.request.route_url('products.view', uuid=row.product_uuid)
            return tags.link_to(text, url)
        return text

    def get_execute_success_url(self, batch, result, **kwargs):
        if kwargs['action'] == 'make_inventory_batch':
            return self.request.route_url('batch.inventory.view', uuid=result.uuid)
        elif kwargs['action'] == 'make_label_batch':
            return self.request.route_url('labels.batch.view', uuid=result.uuid)
        return super(HandheldBatchView, self).get_execute_success_url(batch)

    def get_execute_results_success_url(self, result, **kwargs):
        if result is True:
            # no batches were actually executed
            return self.get_index_url()
        batch = result
        return self.get_execute_success_url(batch, result, **kwargs)


def includeme(config):
    HandheldBatchView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Person Views
"""

from __future__ import unicode_literals, absolute_import

import six
import sqlalchemy as sa
from sqlalchemy import orm

from rattail.db import model, api

import colander
from pyramid.httpexceptions import HTTPFound, HTTPNotFound
from webhelpers2.html import HTML, tags

from tailbone import forms, grids
from tailbone.views import MasterView, AutocompleteView


class PeopleView(MasterView):
    """
    Master view for the Person class.
    """
    model_class = model.Person
    model_title_plural = "People"
    route_prefix = 'people'
    has_versions = True
    supports_mobile = True
    manage_notes_from_profile_view = False

    grid_columns = [
        'display_name',
        'first_name',
        'last_name',
        'phone',
        'email',
    ]

    form_fields = [
        'first_name',
        'middle_name',
        'last_name',
        'display_name',
        'phone',
        'email',
        'address',
        'employee',
        'customers',
        'users',
    ]

    mobile_form_fields = [
        'first_name',
        'middle_name',
        'last_name',
        'display_name',
        'phone',
        'email',
        'address',
        'employee',
        'customers',
        'users',
    ]

    def configure_grid(self, g):
        super(PeopleView, self).configure_grid(g)

        g.joiners['email'] = lambda q: q.outerjoin(model.PersonEmailAddress, sa.and_(
            model.PersonEmailAddress.parent_uuid == model.Person.uuid,
            model.PersonEmailAddress.preference == 1))
        g.joiners['phone'] = lambda q: q.outerjoin(model.PersonPhoneNumber, sa.and_(
            model.PersonPhoneNumber.parent_uuid == model.Person.uuid,
            model.PersonPhoneNumber.preference == 1))

        g.filters['email'] = g.make_filter('email', model.PersonEmailAddress.address)
        g.set_filter('phone', model.PersonPhoneNumber.number,
                     factory=grids.filters.AlchemyPhoneNumberFilter)

        g.joiners['customer_id'] = lambda q: q.outerjoin(model.CustomerPerson).outerjoin(model.Customer)
        g.filters['customer_id'] = g.make_filter('customer_id', model.Customer.id)

        g.filters['first_name'].default_active = True
        g.filters['first_name'].default_verb = 'contains'

        g.filters['last_name'].default_active = True
        g.filters['last_name'].default_verb = 'contains'

        g.sorters['email'] = lambda q, d: q.order_by(getattr(model.PersonEmailAddress.address, d)())
        g.sorters['phone'] = lambda q, d: q.order_by(getattr(model.PersonPhoneNumber.number, d)())

        g.set_sort_defaults('display_name')

        g.set_label('display_name', "Full Name")
        g.set_label('phone', "Phone Number")
        g.set_label('email', "Email Address")
        g.set_label('customer_id', "Customer ID")

        g.set_link('display_name')
        g.set_link('first_name')
        g.set_link('last_name')

    def get_instance(self):
        # TODO: I don't recall why this fallback check for a vendor contact
        # exists here, but leaving it intact for now.
        key = self.request.matchdict['uuid']
        instance = self.Session.query(model.Person).get(key)
        if instance:
            return instance
        instance = self.Session.query(model.VendorContact).get(key)
        if instance:
            return instance.person
        raise HTTPNotFound

    def editable_instance(self, person):
        if self.rattail_config.demo():
            return not bool(person.user and person.user.username == 'chuck')
        return True

    def deletable_instance(self, person):
        if self.rattail_config.demo():
            return not bool(person.user and person.user.username == 'chuck')
        return True

    def configure_common_form(self, f):
        super(PeopleView, self).configure_common_form(f)

        f.set_label('display_name', "Full Name")

        f.set_readonly('phone')
        f.set_label('phone', "Phone Number")

        f.set_readonly('email')
        f.set_label('email', "Email Address")

        f.set_readonly('address')
        f.set_label('address', "Mailing Address")

        # employee
        if self.creating:
            f.remove_field('employee')
        else:
            f.set_readonly('employee')
            f.set_renderer('employee', self.render_employee)

        # customers
        if self.creating:
            f.remove_field('customers')
        else:
            f.set_readonly('customers')
            f.set_renderer('customers', self.render_customers)

        # users
        if self.creating:
            f.remove_field('users')
        else:
            f.set_readonly('users')
            f.set_renderer('users', self.render_users)

    def render_employee(self, person, field):
        employee = person.employee
        if not employee:
            return ""
        text = six.text_type(employee)
        url = self.request.route_url('employees.view', uuid=employee.uuid)
        return tags.link_to(text, url)

    def render_customers(self, person, field):
        customers = person._customers
        if not customers:
            return ""
        items = []
        for customer in customers:
            customer = customer.customer
            text = six.text_type(customer)
            if customer.id:
                text = "({}) {}".format(customer.id, text)
            elif customer.number:
                text = "({}) {}".format(customer.number, text)
            route = '{}customers.view'.format('mobile.' if self.mobile else '')
            url = self.request.route_url(route, uuid=customer.uuid)
            items.append(HTML.tag('li', c=[tags.link_to(text, url)]))
        return HTML.tag('ul', c=items)

    def render_users(self, person, field):
        use_buefy = self.get_use_buefy()
        users = person.users
        items = []
        for user in users:
            text = user.username
            url = self.request.route_url('users.view', uuid=user.uuid)
            items.append(HTML.tag('li', c=[tags.link_to(text, url)]))
        if items:
            return HTML.tag('ul', c=items)
        elif self.viewing and self.request.has_perm('users.create'):
            if use_buefy:
                return HTML.tag('button', type='button', class_='button is-primary',
                                id='make-user', c="Make User")
            else:
                return HTML.tag('button', type='button', id='make-user', c="Make User")
        else:
            return ""

    def get_version_child_classes(self):
        return [
            (model.PersonPhoneNumber, 'parent_uuid'),
            (model.PersonEmailAddress, 'parent_uuid'),
            (model.PersonMailingAddress, 'parent_uuid'),
            (model.Employee, 'person_uuid'),
            (model.CustomerPerson, 'person_uuid'),
            (model.VendorContact, 'person_uuid'),
        ]

    def view_profile(self):
        """
        View which exposes the "full profile" for a given person, i.e. all
        related customer, employee, user info etc.
        """
        self.viewing = True
        person = self.get_instance()
        employee = person.employee
        context = {
            'person': person,
            'instance': person,
            'instance_title': self.get_instance_title(person),
            'employee': employee,
            'employee_history': employee.get_current_history() if employee else None,
        }
        use_buefy = self.get_use_buefy()
        template = 'view_profile_buefy' if use_buefy else 'view_profile'
        return self.render_to_response(template, context)

    def make_note_form(self, mode, person):
        schema = NoteSchema().bind(session=self.Session(),
                                   person_uuid=person.uuid)
        if mode == 'create':
            del schema['uuid']
        form = forms.Form(schema=schema, request=self.request)
        return form

    def profile_add_note(self):
        person = self.get_instance()
        form = self.make_note_form('create', person)
        if form.validate(newstyle=True):
            note = self.create_note(person, form)
            self.Session.flush()
            return self.profile_add_note_success(note)
        else:
            return self.profile_add_note_failure(person, form)

    def create_note(self, person, form):
        note = model.PersonNote()
        note.type = form.validated['note_type']
        note.subject = form.validated['note_subject']
        note.text = form.validated['note_text']
        note.created_by = self.request.user
        person.notes.append(note)
        return note

    def profile_add_note_success(self, note):
        return self.redirect(self.get_action_url('view_profile', person))

    def profile_add_note_failure(self, person, form):
        return self.redirect(self.get_action_url('view_profile', person))

    def profile_edit_note(self):
        person = self.get_instance()
        form = self.make_note_form('edit', person)
        if form.validate(newstyle=True):
            note = self.update_note(person, form)
            self.Session.flush()
            return self.profile_edit_note_success(note)
        else:
            return self.profile_edit_note_failure(person, form)

    def update_note(self, person, form):
        note = self.Session.query(model.PersonNote).get(form.validated['uuid'])
        note.subject = form.validated['note_subject']
        note.text = form.validated['note_text']
        return note

    def profile_edit_note_success(self, note):
        return self.redirect(self.get_action_url('view_profile', person))

    def profile_edit_note_failure(self, person, form):
        return self.redirect(self.get_action_url('view_profile', person))

    def profile_delete_note(self):
        person = self.get_instance()
        form = self.make_note_form('delete', person)
        if form.validate(newstyle=True):
            self.delete_note(person, form)
            self.Session.flush()
            return self.profile_delete_note_success(person)
        else:
            return self.profile_delete_note_failure(person, form)

    def delete_note(self, person, form):
        note = self.Session.query(model.PersonNote).get(form.validated['uuid'])
        self.Session.delete(note)

    def profile_delete_note_success(self, person):
        return self.redirect(self.get_action_url('view_profile', person))

    def profile_delete_note_failure(self, person, form):
        return self.redirect(self.get_action_url('view_profile', person))

    def make_user(self):
        uuid = self.request.POST['person_uuid']
        person = self.Session.query(model.Person).get(uuid)
        if not person:
            return self.notfound()
        if person.users:
            raise RuntimeError("person {} already has {} user accounts: ".format(
                person.uuid, len(person.users), person))
        user = model.User()
        user.username = api.make_username(person)
        user.person = person
        user.active = False
        self.Session.add(user)
        self.Session.flush()
        self.request.session.flash("User has been created: {}".format(user.username))
        return self.redirect(self.request.route_url('users.view', uuid=user.uuid))

    @classmethod
    def defaults(cls, config):
        cls._people_defaults(config)
        cls._defaults(config)

    @classmethod
    def _people_defaults(cls, config):
        permission_prefix = cls.get_permission_prefix()
        route_prefix = cls.get_route_prefix()
        url_prefix = cls.get_url_prefix()
        model_key = cls.get_model_key()
        model_title = cls.get_model_title()

        # view profile
        config.add_tailbone_permission(permission_prefix, '{}.view_profile'.format(permission_prefix),
                                       "View full \"profile\" for {}".format(model_title))
        config.add_route('{}.view_profile'.format(route_prefix), '{}/{{{}}}/profile'.format(url_prefix, model_key),
                         request_method='GET')
        config.add_view(cls, attr='view_profile', route_name='{}.view_profile'.format(route_prefix),
                        permission='{}.view_profile'.format(permission_prefix))

        # manage notes from profile view
        if cls.manage_notes_from_profile_view:

            # add note
            config.add_tailbone_permission(permission_prefix, '{}.profile_add_note'.format(permission_prefix),
                                           "Add new {} note from profile view".format(model_title))
            config.add_route('{}.profile_add_note'.format(route_prefix), '{}/{{{}}}/profile/new-note'.format(url_prefix, model_key),
                             request_method='POST')
            config.add_view(cls, attr='profile_add_note', route_name='{}.profile_add_note'.format(route_prefix),
                            permission='{}.profile_add_note'.format(permission_prefix))

            # edit note
            config.add_tailbone_permission(permission_prefix, '{}.profile_edit_note'.format(permission_prefix),
                                           "Edit existing {} note from profile view".format(model_title))
            config.add_route('{}.profile_edit_note'.format(route_prefix), '{}/{{{}}}/profile/edit-note'.format(url_prefix, model_key),
                             request_method='POST')
            config.add_view(cls, attr='profile_edit_note', route_name='{}.profile_edit_note'.format(route_prefix),
                            permission='{}.profile_edit_note'.format(permission_prefix))

            # delete note
            config.add_tailbone_permission(permission_prefix, '{}.profile_delete_note'.format(permission_prefix),
                                           "Delete existing {} note from profile view".format(model_title))
            config.add_route('{}.profile_delete_note'.format(route_prefix), '{}/{{{}}}/profile/delete-note'.format(url_prefix, model_key),
                             request_method='POST')
            config.add_view(cls, attr='profile_delete_note', route_name='{}.profile_delete_note'.format(route_prefix),
                            permission='{}.profile_delete_note'.format(permission_prefix))

        # make user for person
        config.add_route('{}.make_user'.format(route_prefix), '{}/make-user'.format(url_prefix),
                         request_method='POST')
        config.add_view(cls, attr='make_user', route_name='{}.make_user'.format(route_prefix),
                        permission='users.create')


class PeopleAutocomplete(AutocompleteView):

    mapped_class = model.Person
    fieldname = 'display_name'


class PeopleEmployeesAutocomplete(PeopleAutocomplete):
    """
    Autocomplete view for the Person model, but restricted to return only
    results for people who are employees.
    """

    def filter_query(self, q):
        return q.join(model.Employee)


class PersonNoteView(MasterView):
    """
    Master view for the PersonNote class.
    """
    model_class = model.PersonNote
    route_prefix = 'person_notes'
    url_prefix = '/people/notes'
    has_versions = True

    grid_columns = [
        'person',
        'type',
        'subject',
        'created',
        'created_by',
    ]

    form_fields = [
        'person',
        'type',
        'subject',
        'text',
        'created',
        'created_by',
    ]

    def get_instance_title(self, note):
        return note.subject or "(no subject)"

    def configure_grid(self, g):
        super(PersonNoteView, self).configure_grid(g)

        # person
        g.set_joiner('person', lambda q: q.join(model.Person,
                                                model.Person.uuid == model.PersonNote.parent_uuid))
        g.set_sorter('person', model.Person.display_name)
        g.set_filter('person', model.Person.display_name, label="Person Name")

        # created_by
        CreatorPerson = orm.aliased(model.Person)
        g.set_joiner('created_by', lambda q: q.join(model.User).outerjoin(CreatorPerson,
                                                                          CreatorPerson.uuid == model.User.person_uuid))
        g.set_sorter('created_by', CreatorPerson.display_name)

        g.set_sort_defaults('created', 'desc')

        g.set_link('person')
        g.set_link('subject')
        g.set_link('created')

    def configure_form(self, f):
        super(PersonNoteView, self).configure_form(f)

        # person
        f.set_readonly('person')
        f.set_renderer('person', self.render_person)

        # created
        f.set_readonly('created')

        # created_by
        f.set_readonly('created_by')
        f.set_renderer('created_by', self.render_user)


@colander.deferred
def valid_note_uuid(node, kw):
    session = kw['session']
    person_uuid = kw['person_uuid']
    def validate(node, value):
        note = session.query(model.PersonNote).get(value)
        if not note:
            raise colander.Invalid(node, "Note not found")
        if note.person.uuid != person_uuid:
            raise colander.Invalid(node, "Note is for the wrong person")
        return note.uuid
    return validate


class NoteSchema(colander.Schema):

    uuid = colander.SchemaNode(colander.String(),
                               validator=valid_note_uuid)

    note_type = colander.SchemaNode(colander.String())

    note_subject = colander.SchemaNode(colander.String(), missing='')

    note_text = colander.SchemaNode(colander.String(), missing='')


def includeme(config):

    # autocomplete
    config.add_route('people.autocomplete', '/people/autocomplete')
    config.add_view(PeopleAutocomplete, route_name='people.autocomplete',
                    renderer='json', permission='people.list')
    config.add_route('people.autocomplete.employees', '/people/autocomplete/employees')
    config.add_view(PeopleEmployeesAutocomplete, route_name='people.autocomplete.employees',
                    renderer='json', permission='people.list')

    PeopleView.defaults(config)
    PersonNoteView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master class for generic export history views
"""

from __future__ import unicode_literals, absolute_import

import os

import six

from rattail.db import model

from pyramid.response import FileResponse
from webhelpers2.html import HTML, tags

from tailbone import forms
from tailbone.views import MasterView


class ExportMasterView(MasterView):
    """
    Master class for generic export history views
    """
    creatable = False
    editable = False
    export_has_file = False

    grid_columns = [
        'id',
        'created',
        'created_by',
        'record_count',
    ]

    form_fields = [
        'id',
        'created',
        'created_by',
        'record_count',
    ]

    def get_export_key(self):
        if hasattr(self, 'export_key'):
            return self.export_key

    def get_file_path(self, export, makedirs=False):
        return self.rattail_config.export_filepath(self.export_key,
                                                   export.uuid,
                                                   export.filename,
                                                   makedirs=makedirs)

    def configure_grid(self, g):
        super(ExportMasterView, self).configure_grid(g)

        g.joiners['created_by'] = lambda q: q.join(model.User)
        g.sorters['created_by'] = g.make_sorter(model.User.username)
        g.filters['created_by'] = g.make_filter('created_by', model.User.username)
        g.set_sort_defaults('created', 'desc')

        g.set_renderer('id', self.render_id)

        g.set_label('id', "ID")
        g.set_label('created_by', "Created by")

        g.set_link('id')

    def render_id(self, export, field):
        return export.id_str

    def configure_form(self, f):
        super(ExportMasterView, self).configure_form(f)
        export = f.model_instance

        # NOTE: we try to handle the 'creating' scenario even though this class
        # doesn't officially support that; just in case a subclass does want to

        # id
        if self.creating:
            f.remove_field('id')
        else:
            f.set_readonly('id')
            f.set_renderer('id', self.render_id)
            f.set_label('id', "ID")

        # created
        if self.creating:
            f.remove_field('created')
        else:
            f.set_readonly('created')
            f.set_type('created', 'datetime')

        # created_by
        if self.creating:
            f.remove_field('created_by')
        else:
            f.set_readonly('created_by')
            f.set_renderer('created_by', self.render_created_by)
            f.set_label('created_by', "Created by")

        # record_count
        if self.creating:
            f.remove_field('record_count')
        else:
            f.set_readonly('record_count')

        # download
        if self.export_has_file and self.viewing:
            # download = forms.renderers.FileFieldRenderer.new(
            #     self, storage_path=self.rattail_config.export_filedir(self.export_key),
            #     file_path=self.get_file_path(fs.model), download_url=self.get_download_url)
            f.append('download')
            f.set_renderer('download', self.render_download)

    def render_download(self, export, field):
        path = self.get_file_path(export)
        text = "{} ({})".format(export.filename, self.readable_size(path))
        url = self.request.route_url('{}.download'.format(self.get_route_prefix()), uuid=export.uuid)
        return tags.link_to(text, url)

    def render_created_by(self, export, field):
        user = export.created_by
        if not user:
            return ""
        text = six.text_type(user)
        if self.request.has_perm('users.view'):
            url = self.request.route_url('users.view', uuid=user.uuid)
            return tags.link_to(text, url)
        return text

    def get_download_url(self, filename):
        uuid = self.request.matchdict['uuid']
        return self.request.route_url('{}.download'.format(self.get_route_prefix()), uuid=uuid)

    def download(self):
        """
        View for downloading the export file.
        """
        export = self.get_instance()
        path = self.get_file_path(export)
        response = FileResponse(path, request=self.request)
        if six.PY3:
            response.headers['Content-Length'] = str(os.path.getsize(path))
            response.headers['Content-Disposition'] = 'attachment; filename="{}"'.format(export.filename)
        else:
            response.headers[b'Content-Length'] = six.binary_type(os.path.getsize(path))
            response.headers[b'Content-Disposition'] = b'attachment; filename="{}"'.format(export.filename)
        return response

    def delete_instance(self, export):
        """
        Delete the export file also, if it exists.
        """
        if self.export_has_file:
            path = self.get_file_path(export)
            if os.path.exists(path):
                os.remove(path)
                os.rmdir(os.path.dirname(path))
        super(ExportMasterView, self).delete_instance(export)

    @classmethod
    def defaults(cls, config):
        """
        Provide default configuration for a master view.
        """
        cls._defaults(config)
        cls._export_defaults(config)

    @classmethod
    def _export_defaults(cls, config):
        route_prefix = cls.get_route_prefix()
        url_prefix = cls.get_url_prefix()
        permission_prefix = cls.get_permission_prefix()
        model_key = cls.get_model_key()
        model_title = cls.get_model_title()

        # download export file
        if cls.export_has_file:
            config.add_route('{}.download'.format(route_prefix), '{}/{{{}}}/download'.format(url_prefix, model_key))
            config.add_view(cls, attr='download', route_name='{}.download'.format(route_prefix),
                            permission='{}.download'.format(permission_prefix))
            config.add_tailbone_permission(permission_prefix, '{}.download'.format(permission_prefix),
                                           "Download {} data file".format(model_title))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Customer order item views
"""

from __future__ import unicode_literals, absolute_import

import datetime

import six
from sqlalchemy import orm

from rattail.db import model
from rattail.time import localtime

from webhelpers2.html import tags

from tailbone.views import MasterView
from tailbone.util import raw_datetime


class CustomerOrderItemsView(MasterView):
    """
    Master view for customer order items
    """
    model_class = model.CustomerOrderItem
    route_prefix = 'custorders.items'
    url_prefix = '/custorders/items'
    creatable = False
    editable = False
    deletable = False

    grid_columns = [
        'person',
        'product_brand',
        'product_description',
        'product_size',
        'case_quantity',
        'cases_ordered',
        'units_ordered',
        'order_created',
        'status_code',
    ]

    has_rows = True
    model_row_class = model.CustomerOrderItemEvent
    rows_title = "Event History"
    rows_filterable = False
    rows_sortable = False
    rows_pageable = False
    rows_viewable = False

    row_grid_columns = [
        'occurred',
        'type_code',
        'user',
        'note',
    ]

    form_fields = [
        'person',
        'product',
        'product_brand',
        'product_description',
        'product_size',
        'case_quantity',
        'cases_ordered',
        'units_ordered',
        'unit_price',
        'total_price',
        'paid_amount',
        'status_code',
    ]

    def query(self, session):
        return session.query(model.CustomerOrderItem)\
                      .join(model.CustomerOrder)\
                      .options(orm.joinedload(model.CustomerOrderItem.order)\
                               .joinedload(model.CustomerOrder.person))

    def configure_grid(self, g):
        super(CustomerOrderItemsView, self).configure_grid(g)

        g.set_joiner('person', lambda q: q.outerjoin(model.Person))

        g.filters['person'] = g.make_filter('person', model.Person.display_name,
                                            default_active=True, default_verb='contains')

        g.set_sorter('person', model.Person.display_name)
        g.set_sorter('order_created', model.CustomerOrder.created)

        g.set_sort_defaults('order_created', 'desc')

        g.set_type('case_quantity', 'quantity')
        g.set_type('cases_ordered', 'quantity')
        g.set_type('units_ordered', 'quantity')
        g.set_type('total_price', 'currency')

        g.set_renderer('person', self.render_person)
        g.set_renderer('order_created', self.render_order_created)

        g.set_label('person', "Person Name")
        g.set_label('product_brand', "Brand")
        g.set_label('product_description', "Description")
        g.set_label('product_size', "Size")
        g.set_label('status_code', "Status")

    def render_person(self, item, column):
        return item.order.person

    def render_order_created(self, item, column):
        value = localtime(self.rattail_config, item.order.created, from_utc=True)
        return raw_datetime(self.rattail_config, value)

    def configure_form(self, f):
        super(CustomerOrderItemsView, self).configure_form(f)

        # order
        f.set_renderer('order', self.render_order)

        # product
        f.set_renderer('product', self.render_product)

        # product uom
        f.set_enum('product_unit_of_measure', self.enum.UNIT_OF_MEASURE)

        # quantity fields
        f.set_type('case_quantity', 'quantity')
        f.set_type('cases_ordered', 'quantity')
        f.set_type('units_ordered', 'quantity')

        # currency fields
        f.set_type('unit_price', 'currency')
        f.set_type('total_price', 'currency')
        f.set_type('paid_amount', 'currency')

        # person
        f.set_renderer('person', self.render_person)

        # label overrides
        f.set_label('status_code', "Status")

    def render_order(self, item, field):
        order = item.order
        if not order:
            return ""
        text = six.text_type(order)
        url = self.request.route_url('custorders.view', uuid=order.uuid)
        return tags.link_to(text, url)

    def get_row_data(self, item):
        return self.Session.query(model.CustomerOrderItemEvent)\
                           .filter(model.CustomerOrderItemEvent.item == item)\
                           .order_by(model.CustomerOrderItemEvent.occurred,
                                     model.CustomerOrderItemEvent.type_code)

    def configure_row_grid(self, g):
        super(CustomerOrderItemsView, self).configure_row_grid(g)
        g.set_label('occurred', "When")
        g.set_label('type_code', "What") # TODO: enum renderer
        g.set_label('user', "Who")
        g.set_label('note', "Notes")


def includeme(config):
    CustomerOrderItemsView.defaults(config)
